"""
This module contains the AseDBDataset class and its dependencies.
It is extracted from the fairchem codebase and adapted to remove dependencies on fairchem.

Original code copyright:
Copyright (c) Meta, Inc. and its affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

import bisect
import logging
import numbers
import os
import zlib
from abc import ABC, abstractmethod

try:
    from functools import cache, cached_property
except ImportError:
    from functools import cached_property, lru_cache

    cache = lru_cache(maxsize=None)
from glob import glob
from pathlib import Path
from typing import Any, Callable, TypeVar

import ase
import ase.db.core
import ase.db.row
import ase.io
import lmdb
import numpy as np
import orjson
import torch

# Type variable for generic dataset return type
T_co = TypeVar("T_co", covariant=True)


def _decode_ndarrays(obj):
    """Recursively turn {"__ndarray__": [...] } blobs back into NumPy arrays."""
    if isinstance(obj, dict):
        if "__ndarray__" in obj:
            shape, dtype, flat = obj["__ndarray__"]
            return np.asarray(flat, dtype=dtype).reshape(shape)
        # recurse into dict values
        return {k: _decode_ndarrays(v) for k, v in obj.items()}
    if isinstance(obj, list):
        return [_decode_ndarrays(v) for v in obj]
    if isinstance(obj, tuple):
        return tuple(_decode_ndarrays(v) for v in obj)
    return obj  # everything else is left untouched


def rename_data_object_keys(data_object, key_mapping: dict[str, str | list[str]]):
    """Rename data object keys

    Args:
        data_object: data object
        key_mapping: dictionary specifying keys to rename and new names {prev_key: new_key}

        new_key can be a list of new keys, for example,
        prev_key: energy
        new_key: [common_energy, oc20_energy]

        This is currently required when we use a single target/label for multiple tasks
    """
    for _property in key_mapping:
        # catch for test data not containing labels
        if _property in data_object:
            list_of_new_keys = key_mapping[_property]
            if isinstance(list_of_new_keys, str):
                list_of_new_keys = [list_of_new_keys]
            for new_property in list_of_new_keys:
                if new_property == _property:
                    continue
                assert new_property not in data_object
                data_object[new_property] = data_object[_property]
            if _property not in list_of_new_keys:
                del data_object[_property]
    return data_object


def apply_one_tags(
    atoms: ase.Atoms, skip_if_nonzero: bool = True, skip_always: bool = False
):
    """
    This function will apply tags of 1 to an ASE atoms object.
    It is used as an atoms_transform in the datasets contained in this file.

    Certain models will treat atoms differently depending on their tags.
    For example, GemNet-OC by default will only compute triplet and quadruplet interactions
    for atoms with non-zero tags. This model throws an error if there are no tagged atoms.
    For this reason, the default behavior is to tag atoms in structures with no tags.

    args:
        skip_if_nonzero (bool): If at least one atom has a nonzero tag, do not tag any atoms

        skip_always (bool): Do not apply any tags. This arg exists so that this function can be disabled
                without needing to pass a callable (which is currently difficult to do with main.py)
    """
    if skip_always:
        return atoms

    if np.all(atoms.get_tags() == 0) or not skip_if_nonzero:
        atoms.set_tags(np.ones(len(atoms)))

    return atoms


class UnsupportedDatasetError(ValueError):
    pass


class BaseDataset(ABC):
    """Base Dataset class for all ASE datasets."""

    def __init__(self, config: dict):
        """Initialize

        Args:
            config (dict): dataset configuration
        """
        self.config = config
        self.paths = []

        if "src" in self.config:
            if isinstance(config["src"], str):
                self.paths = [Path(self.config["src"])]
            else:
                self.paths = tuple(Path(path) for path in sorted(config["src"]))

        self.lin_ref = None
        if self.config.get("lin_ref", False):
            lin_ref = torch.tensor(
                np.load(self.config["lin_ref"], allow_pickle=True)["coeff"]
            )
            self.lin_ref = torch.nn.Parameter(lin_ref, requires_grad=False)

    def __len__(self) -> int:
        return self.num_samples

    def metadata_hasattr(self, attr) -> bool:
        return attr in self._metadata

    @cached_property
    def indices(self):
        return np.arange(self.num_samples, dtype=int)

    @cached_property
    def _metadata(self) -> dict[str, np.ndarray]:
        # logic to read metadata file here
        metadata_npzs = []
        if self.config.get("metadata_path", None) is not None:
            metadata_npzs.append(
                np.load(self.config["metadata_path"], allow_pickle=True)
            )

        else:
            for path in self.paths:
                if path.is_file():
                    metadata_file = path.parent / "metadata.npz"
                else:
                    metadata_file = path / "metadata.npz"
                if metadata_file.is_file():
                    metadata_npzs.append(np.load(metadata_file, allow_pickle=True))

        if len(metadata_npzs) == 0:
            logging.warning(
                f"Could not find dataset metadata.npz files in '{self.paths}'"
            )
            return {}

        metadata = {
            field: np.concatenate([metadata[field] for metadata in metadata_npzs])
            for field in metadata_npzs[0]
        }

        assert np.issubdtype(
            metadata["natoms"].dtype, np.integer
        ), f"Metadata natoms must be an integer type! not {metadata['natoms'].dtype}"
        assert metadata["natoms"].shape[0] == len(
            self
        ), "Loaded metadata and dataset size mismatch."

        return metadata

    def get_metadata(self, attr, idx):
        if attr in self._metadata:
            metadata_attr = self._metadata[attr]
            if isinstance(idx, list):
                return [metadata_attr[_idx] for _idx in idx]
            return metadata_attr[idx]
        return None


class Subset(BaseDataset):
    """A subset that also takes metadata if given."""

    def __init__(
        self,
        dataset: BaseDataset,
        indices: list[int],
        metadata: dict[str, np.ndarray],
    ) -> None:
        super().__init__(dataset.config)
        self.dataset = dataset
        self.metadata = metadata
        self.indices = indices
        self.num_samples = len(indices)
        self.config = dataset.config

    @cached_property
    def _metadata(self) -> dict[str, np.ndarray]:
        return self.dataset._metadata  # pylint: disable=protected-access

    def get_metadata(self, attr, idx):
        if isinstance(idx, list):
            return self.dataset.get_metadata(attr, [[self.indices[i] for i in idx]])
        return self.dataset.get_metadata(attr, self.indices[idx])


class LMDBDatabase(ase.db.core.Database):
    """
    This module is modified from the ASE db json backend
    and is thus licensed under the corresponding LGPL2.1 license.

    The ASE notice for the LGPL2.1 license is available here:
    https://gitlab.com/ase/ase/-/blob/master/LICENSE
    """

    def __init__(  # pylint: disable=keyword-arg-before-vararg
        self,
        filename: str | Path | None = None,
        create_indices: bool = True,
        use_lock_file: bool = False,
        serial: bool = False,
        readonly: bool = False,  # Moved after *args to make it keyword-only
        *args,
        **kwargs,
    ) -> None:
        """
        For the most part, this is identical to the standard ase db initiation
        arguments, except that we add a readonly flag.
        """
        super().__init__(
            Path(filename),
            create_indices,
            use_lock_file,
            serial,
            *args,
            **kwargs,
        )

        # Add a readonly mode for when we're only training
        # to make sure there's no parallel locks
        self.readonly = readonly

        if self.readonly:
            # Open a new env
            self.env = lmdb.open(
                str(self.filename),
                subdir=False,
                meminit=False,
                map_async=True,
                readonly=True,
                lock=False,
            )

            # Open a transaction and keep it open for fast read/writes!
            self.txn = self.env.begin(write=False)

        else:
            # Open a new env with write access
            self.env = lmdb.open(
                str(self.filename),
                map_size=1099511627776 * 2,
                subdir=False,
                meminit=False,
                map_async=True,
            )

            self.txn = self.env.begin(write=True)

        # Load all ids based on keys in the DB.
        self.ids = []
        self.deleted_ids = []
        self._load_ids()

    def __enter__(self) -> "LMDBDatabase":
        return self

    def __exit__(self, exc_type, exc_value, tb) -> None:
        self.close()

    def close(self) -> None:
        # Close the lmdb environment and transaction
        self.txn.commit()
        self.env.close()

    def _write(
        self,
        atoms: ase.Atoms | ase.db.row.AtomsRow,
        key_value_pairs: dict,
        data: dict | None,
        id: int | None = None,  # pylint: disable=redefined-builtin
    ) -> None:

        # 1) dump the entire atoms.info dict into key_value_pairs
        key_value_pairs = dict(key_value_pairs or {}, **atoms.info)
        scalar_types = (numbers.Real, str, bool, np.bool_)
        key_value_pairs = {
            k: v
            for k, v in (key_value_pairs or {}).items()
            if isinstance(v, scalar_types)
        }

        if data is None:
            data = {}
        for k, v in atoms.info.items():
            if isinstance(v, scalar_types):
                key_value_pairs[k] = v
            else:
                # If the value is not serializable, we store it in data
                data.setdefault("__info__", {})[k] = v
        arrays_to_dump = {}
        for name, arr in atoms.arrays.items():
            if name not in (
                "numbers",
                "positions",
                "tags",
                "momenta",
                "masses",
                "charges",
                "magmoms",
                "velocities",
            ):
                arrays_to_dump[name] = arr
        if arrays_to_dump:
            data.setdefault("__arrays__", {}).update(arrays_to_dump)

        # 3) also save all extra calculator results (if any)
        if hasattr(atoms, "calc") and getattr(atoms.calc, "results", None):
            for k, v in atoms.calc.results.items():
                if k in ("energy", "forces", "stress", "free_energy"):
                    continue  # ASE already stores these
            data.setdefault(k, v)
        # Call parent method with the original parameter name
        super()._write(atoms, key_value_pairs, data)

        mtime = ase.db.core.now()

        if isinstance(atoms, ase.db.row.AtomsRow):
            row = atoms
        else:
            row = ase.db.row.AtomsRow(atoms)
            row.ctime = mtime
            row.user = os.getenv("USER")

        dct = {}
        for key in row.__dict__:
            # Use getattr to avoid accessing protected member directly
            if key[0] == "_" or key == "id" or key in getattr(row, "_keys", []):
                continue
            dct[key] = row[key]

        dct["mtime"] = mtime

        if key_value_pairs:
            dct["key_value_pairs"] = key_value_pairs

        if data:
            dct["data"] = data

        constraints = row.get("constraints")
        if constraints:
            dct["constraints"] = [constraint.todict() for constraint in constraints]

        # json doesn't like Cell objects, so make it an array
        dct["cell"] = np.asarray(dct["cell"])

        if id is None:
            id = self._nextid
            nextid = id + 1
        else:
            data = self.txn.get(f"{id}".encode("ascii"))
            assert data is not None

        # Add the new entry
        self.txn.put(
            f"{id}".encode("ascii"),
            zlib.compress(orjson.dumps(dct, option=orjson.OPT_SERIALIZE_NUMPY)),
        )
        # only append if idx is not in ids
        if id not in self.ids:
            self.ids.append(id)
            self.txn.put(
                "nextid".encode("ascii"),
                zlib.compress(orjson.dumps(nextid, option=orjson.OPT_SERIALIZE_NUMPY)),
            )
        # check if id is in removed ids and remove accordingly
        if id in self.deleted_ids:
            self.deleted_ids.remove(id)
            self._write_deleted_ids()

        return id

    def _update(
        self,
        idx: int,
        key_value_pairs: dict | None = None,
        data: dict | None = None,
    ):
        # hack this to play nicely with ASE code
        row = self._get_row(idx, include_data=True)
        if data is not None or key_value_pairs is not None:
            self._write(
                atoms=row, key_value_pairs=key_value_pairs, data=data, id=idx
            )  # Fixed E1123 by using id=idx

    def _write_deleted_ids(self):
        self.txn.put(
            "deleted_ids".encode("ascii"),
            zlib.compress(
                orjson.dumps(self.deleted_ids, option=orjson.OPT_SERIALIZE_NUMPY)
            ),
        )

    def delete(self, ids: list[int]) -> None:
        for idx in ids:
            self.txn.delete(f"{idx}".encode("ascii"))
            self.ids.remove(idx)

        self.deleted_ids += ids
        self._write_deleted_ids()

    def _get_row(self, idx: int, include_data: bool = True):
        if idx is None:
            assert len(self.ids) == 1
            idx = self.ids[0]
        data = self.txn.get(f"{idx}".encode("ascii"))

        if data is not None:
            dct = orjson.loads(zlib.decompress(data))
        else:
            raise KeyError(f"Id {idx} missing from the database!")

        if not include_data:
            dct.pop("data", None)

        dct["id"] = idx
        return ase.db.row.AtomsRow(dct)

    def _get_row_by_index(self, index: int, include_data: bool = True):
        """Auxiliary function to get the ith entry, rather than a specific id"""
        data = self.txn.get(f"{self.ids[index]}".encode("ascii"))

        if data is not None:
            dct = orjson.loads(zlib.decompress(data))
        else:
            raise KeyError(f"Id {id} missing from the database!")

        if not include_data:
            dct.pop("data", None)

        dct["id"] = id
        return ase.db.row.AtomsRow(dct)

    def _select(
        self,
        keys,
        cmps: list[tuple[str, str, str]],
        explain: bool = False,
        _verbosity: int = 0,  # Unused parameter marked with underscore
        limit: int | None = None,
        offset: int = 0,
        sort: str | None = None,
        include_data: bool = True,
        _columns: str = "all",  # Unused parameter marked with underscore
    ):
        if explain:
            yield {"explain": (0, 0, 0, "scan table")}
            return

        if sort is not None:
            if sort[0] == "-":
                reverse = True
                sort = sort[1:]
            else:
                reverse = False

            rows = []
            missing = []
            for row in self._select(keys, cmps):
                key = row.get(sort)
                if key is None:
                    missing.append((0, row))
                else:
                    rows.append((key, row))

            rows.sort(reverse=reverse, key=lambda x: x[0])
            rows += missing

            if limit:
                rows = rows[offset : offset + limit]
            for _, row in rows:
                yield row
            return

        if not limit:
            limit = -offset - 1

        cmps = [(key, ase.db.core.ops[op], val) for key, op, val in cmps]
        n = 0
        for idx in self.ids:
            if n - offset == limit:
                return
            row = self._get_row(idx, include_data=include_data)

            for key in keys:
                if key not in row:
                    break
            else:
                for key, op, val in cmps:
                    if isinstance(key, int):
                        value = np.equal(row.numbers, key).sum()
                    else:
                        value = row.get(key)
                        if key == "pbc":
                            assert op in [ase.db.core.ops["="], ase.db.core.ops["!="]]
                            value = "".join("FT"[x] for x in value)
                    if value is None or not op(value, val):
                        break
                else:
                    if n >= offset:
                        yield row
                    n += 1

    @property
    def metadata(self):
        """Override abstract metadata method from Database class."""
        return self.db_metadata

    @property
    def db_metadata(self):
        """Load the metadata from the DB if present"""
        if self._metadata is None:
            metadata = self.txn.get("metadata".encode("ascii"))
            if metadata is None:
                self._metadata = {}
            else:
                self._metadata = orjson.loads(zlib.decompress(metadata))

        return self._metadata.copy()

    @db_metadata.setter
    def db_metadata(self, dct):
        self._metadata = dct

        # Put the updated metadata dictionary
        self.txn.put(
            "metadata".encode("ascii"),
            zlib.compress(orjson.dumps(dct, option=orjson.OPT_SERIALIZE_NUMPY)),
        )

    @property
    def _nextid(self):
        """Get the id of the next row to be written"""
        # Get the nextid
        nextid_data = self.txn.get("nextid".encode("ascii"))
        if nextid_data:
            return orjson.loads(zlib.decompress(nextid_data))
        return 1  # Removed unnecessary else (R1705)

    def count(self, selection=None, **kwargs) -> int:
        """Count rows.

        See the select() method for the selection syntax.  Use db.count() or
        len(db) to count all rows.
        """
        if selection is not None:
            n = 0
            for _row in self.select(selection, **kwargs):
                n += 1
            return n
        return len(self.ids)

    def _load_ids(self) -> None:
        """Load ids from the DB

        Since ASE db ids are mostly 1-N integers, but can be missing entries
        if ids have been deleted. To save space and operating under the assumption
        that there will probably not be many deletions in most OCP datasets,
        we just store the deleted ids.
        """
        # Load the deleted ids
        deleted_ids_data = self.txn.get("deleted_ids".encode("ascii"))
        if deleted_ids_data is not None:
            self.deleted_ids = orjson.loads(zlib.decompress(deleted_ids_data))

        # Reconstruct the full id list
        self.ids = [i for i in range(1, self._nextid) if i not in set(self.deleted_ids)]


# Placeholder for AtomsToGraphs class
# This is a minimal implementation without the full functionality
class AtomsToGraphs:
    """Enhanced AtomsToGraphs implementation with proper property handling."""

    def __init__(
        self,
        r_edges=False,
        r_pbc=True,
        r_energy=False,
        r_forces=False,
        r_stress=False,
        r_data_keys=None,
        **kwargs,
    ):
        self.r_edges = r_edges
        self.r_pbc = r_pbc
        self.r_energy = r_energy
        self.r_forces = r_forces
        self.r_stress = r_stress
        self.r_data_keys = r_data_keys or {}
        self.kwargs = kwargs

    def convert(self, atoms, sid=None):
        """
        Convert ASE atoms to graph data format with proper property handling.
        """
        from macer.externals.mace_bundled.tools.torch_geometric.data import Data

        # Create a minimal data object with required properties
        data = Data()

        # Set positions
        data.pos = torch.tensor(atoms.get_positions(), dtype=torch.float)

        # Set atomic numbers
        data.atomic_numbers = torch.tensor(atoms.get_atomic_numbers(), dtype=torch.long)

        # Set cell if available
        if atoms.cell is not None:
            data.cell = torch.tensor(atoms.get_cell(), dtype=torch.float)

        # Set PBC if requested
        if self.r_pbc:
            data.pbc = torch.tensor(atoms.get_pbc(), dtype=torch.bool)

        # Set energy if requested
        if self.r_energy:
            energy = self._get_property(atoms, "energy")
            if energy is not None:
                data.energy = torch.tensor(energy, dtype=torch.float)

        # Set forces if requested
        if self.r_forces:
            forces = self._get_property(atoms, "forces")
            if forces is not None:
                data.forces = torch.tensor(forces, dtype=torch.float)

        # Set stress if requested
        if self.r_stress:
            stress = self._get_property(atoms, "stress")
            if stress is not None:
                data.stress = torch.tensor(stress, dtype=torch.float)

        # Set sid if provided
        if sid is not None:
            data.sid = sid

        return data

    def _get_property(self, atoms, prop_name):
        """Get property from atoms, checking custom names first then standard methods."""
        # Check if we have a custom name for this property
        custom_name = self.r_data_keys.get(prop_name)

        # Try custom name in info dict
        if custom_name and custom_name in atoms.info:
            return atoms.info[custom_name]

        # Try custom name in arrays dict
        if custom_name and custom_name in atoms.arrays:
            return atoms.arrays[custom_name]

        # Try standard name in info dict
        if prop_name in atoms.info:
            return atoms.info[prop_name]

        # Try standard name in arrays dict
        if prop_name in atoms.arrays:
            return atoms.arrays[prop_name]

        # Try standard ASE methods
        method_map = {
            "energy": "get_potential_energy",
            "forces": "get_forces",
            "stress": "get_stress",
        }

        if prop_name in method_map and hasattr(atoms, method_map[prop_name]):
            try:
                method = getattr(atoms, method_map[prop_name])
                return method()
            except (
                AttributeError,
                RuntimeError,
            ) as exc:  # Fixed W0718 by specifying exceptions
                logging.debug(f"Error getting property {prop_name}: {exc}")
                # Removed unnecessary pass (W0107)

        return None


# Placeholder for DataTransforms class
class DataTransforms:
    """Minimal implementation of DataTransforms to satisfy dependencies."""

    def __init__(self, transforms_config=None):
        self.transforms_config = transforms_config or {}

    def __call__(self, data):
        """Apply transforms to data"""
        # No transforms applied in this minimal implementation
        return data


class AseAtomsDataset(BaseDataset, ABC):
    """
    This is an abstract Dataset that includes helpful utilities for turning
    ASE atoms objects into OCP-usable data objects. This should not be instantiated directly
    as get_atoms_object and load_dataset_get_ids are not implemented in this base class.

    Derived classes must add at least two things:
        self.get_atoms_object(id): a function that takes an identifier and returns a corresponding atoms object

        self.load_dataset_get_ids(config: dict): This function is responsible for any initialization/loads
            of the dataset and importantly must return a list of all possible identifiers that can be passed into
            self.get_atoms_object(id)

    Identifiers need not be any particular type.
    """

    def __init__(
        self,
        config: dict,
        atoms_transform: Callable[[ase.Atoms, Any], ase.Atoms] = apply_one_tags,
    ) -> None:
        super().__init__(config)

        a2g_args = config.get("a2g_args", {}) or {}

        # set default to False if not set by user, assuming otf_graph will be used
        if "r_edges" not in a2g_args:
            a2g_args["r_edges"] = False

        # Make sure we always include PBC info in the resulting atoms objects
        a2g_args["r_pbc"] = True
        self.a2g = AtomsToGraphs(**a2g_args)

        self.key_mapping = self.config.get("key_mapping", None)
        self.transforms = DataTransforms(self.config.get("transforms", {}))

        self.atoms_transform = atoms_transform

        if self.config.get("keep_in_memory", False):
            self.__getitem__ = cache(self.__getitem__)

        self.ids = self._load_dataset_get_ids(config)
        self.num_samples = len(self.ids)

        if len(self.ids) == 0:
            raise ValueError(
                rf"No valid ase data found! \n"
                f"Double check that the src path and/or glob search pattern gives ASE compatible data: {config['src']}"
            )

    def __getitem__(self, idx):  # pylint: disable=method-hidden
        # Handle slicing
        if isinstance(idx, slice):
            return [self[i] for i in range(*idx.indices(len(self)))]

        # Get atoms object via derived class method
        atoms = self.get_atoms(self.ids[idx])

        # Transform atoms object
        if self.atoms_transform is not None:
            atoms = self.atoms_transform(
                atoms, **self.config.get("atoms_transform_args", {})
            )

        sid = atoms.info.get("sid", self.ids[idx])
        fid = atoms.info.get("fid", torch.tensor([0]))

        # Convert to data object
        data_object = self.a2g.convert(atoms, sid)
        data_object.fid = fid
        data_object.natoms = len(atoms)

        # apply linear reference
        if self.a2g.r_energy is True and self.lin_ref is not None:
            data_object.energy -= sum(self.lin_ref[data_object.atomic_numbers.long()])

        # Transform data object
        data_object = self.transforms(data_object)

        if self.key_mapping is not None:
            data_object = rename_data_object_keys(data_object, self.key_mapping)

        if self.config.get("include_relaxed_energy", False):
            data_object.energy_relaxed = self.get_relaxed_energy(self.ids[idx])

        return data_object

    @abstractmethod
    def get_atoms(self, idx: str | int) -> ase.Atoms:
        # This function should return an ASE atoms object.
        raise NotImplementedError(
            "Returns an ASE atoms object. Derived classes should implement this function."
        )

    @abstractmethod
    def _load_dataset_get_ids(self, config):
        # This function should return a list of ids that can be used to index into the database
        raise NotImplementedError(
            "Every ASE dataset needs to declare a function to load the dataset and return a list of ids."
        )

    def get_relaxed_energy(self, identifier):
        raise NotImplementedError(
            "Reading relaxed energy from trajectory or file is not implemented with this dataset. "
            "If relaxed energies are saved with the atoms info dictionary, they can be used by passing the keys in "
            "the r_data_keys argument under a2g_args."
        )

    def get_metadata(self, attr, idx):
        # try the parent method
        metadata = super().get_metadata(attr, idx)
        if metadata is not None:
            return metadata
        # try to resolve it here
        if attr != "natoms":
            return None
        if isinstance(idx, (list, np.ndarray)):
            return np.array([self.get_metadata(attr, i) for i in idx])
        return len(self.get_atoms(idx))


class AseDBDataset(AseAtomsDataset):
    """
    This Dataset connects to an ASE Database, allowing the storage of atoms objects
    with a variety of backends including JSON, SQLite, and database server options.
    """

    def _load_dataset_get_ids(self, config: dict) -> list[int]:
        if isinstance(config["src"], list):
            filepaths = []
            for path in sorted(config["src"]):
                if os.path.isdir(path):
                    filepaths.extend(sorted(glob(f"{path}/*")))
                elif os.path.isfile(path):
                    filepaths.append(path)
                else:
                    raise RuntimeError(f"Error reading dataset in {path}!")
        elif os.path.isfile(config["src"]):
            filepaths = [config["src"]]
        elif os.path.isdir(config["src"]):
            filepaths = sorted(glob(f'{config["src"]}/*'))
        else:
            filepaths = sorted(glob(config["src"]))

        self.dbs = []

        for path in filepaths:
            try:
                self.dbs.append(self.connect_db(path, config.get("connect_args", {})))
            except ValueError:
                logging.debug(
                    f"Tried to connect to {path} but it's not an ASE database!"
                )

        self.select_args = config.get("select_args", {})
        if self.select_args is None:
            self.select_args = {}

        # Get all unique IDs from the databases
        self.db_ids = []
        for db in self.dbs:
            if hasattr(db, "ids") and self.select_args == {}:
                self.db_ids.append(db.ids)
            else:
                # this is the slow alternative
                self.db_ids.append([row.id for row in db.select(**self.select_args)])

        idlens = [len(ids) for ids in self.db_ids]
        self._idlen_cumulative = np.cumsum(idlens).tolist()

        return list(range(sum(idlens)))

    def get_atoms(self, idx: int) -> ase.Atoms:
        """
        Return an `ase.Atoms` object for the dataset entry `idx`, decoding any
        JSON‐encoded ndarrays encountered anywhere in the row.
        """
        # ------------------------------------------------------------------ #
        # 1.  Locate the correct database and row                            #
        # ------------------------------------------------------------------ #
        db_idx = bisect.bisect(self._idlen_cumulative, idx)
        local_idx = idx - self._idlen_cumulative[db_idx - 1] if db_idx else idx
        row = self.get_row_from_db(db_idx, local_idx)

        # ------------------------------------------------------------------ #
        # 2.  Fast path if ASE can already parse the row natively            #
        # ------------------------------------------------------------------ #
        if not (isinstance(row.numbers, dict) and "__ndarray__" in row.numbers):
            atoms = row.toatoms()
        else:
            # -------------------------------------------------------------- #
            # 3.  Decode *everything* that might hide __ndarray__ blobs      #
            # -------------------------------------------------------------- #
            atom_numbers = _decode_ndarrays(row.numbers)
            positions = _decode_ndarrays(row.positions)
            cell = _decode_ndarrays(getattr(row, "cell", None))
            pbc = _decode_ndarrays(getattr(row, "pbc", None))

            atoms = ase.Atoms(
                numbers=atom_numbers,
                positions=positions,
                cell=cell if cell is not None else None,
                pbc=pbc if pbc is not None else None,
            )

        # ------------------------------------------------------------------ #
        # 4.  Row-level dictionaries (data / key_value_pairs) – deep decode  #
        # ------------------------------------------------------------------ #
        data_dict = _decode_ndarrays(row.data) if isinstance(row.data, dict) else {}
        kvp_dict = (
            _decode_ndarrays(row.key_value_pairs)
            if getattr(row, "key_value_pairs", None)
            else {}
        )

        atoms.info.update(data_dict)
        atoms.info.update(kvp_dict)

        # ------------------------------------------------------------------ #
        # 5.  Energy, forces, stress → atoms.calc (decode if needed)         #
        # ------------------------------------------------------------------ #
        calc_kwargs = {}
        for prop in ("energy", "forces", "stress", "free_energy"):
            val = getattr(row, prop, None)
            if val is not None:
                calc_kwargs[prop] = _decode_ndarrays(val)
                atoms.info[prop] = calc_kwargs[prop]

        if calc_kwargs:
            from ase.calculators.singlepoint import SinglePointCalculator

            atoms.calc = SinglePointCalculator(atoms, **calc_kwargs)

        # ------------------------------------------------------------------ #
        # 6.  Extra arrays & info stored under __arrays__ / __info__         #
        # ------------------------------------------------------------------ #
        extra_arrays = data_dict.pop("__arrays__", {})
        for name, arr in extra_arrays.items():
            atoms.new_array(name, np.asarray(arr))  # already decoded above

        extra_info = data_dict.pop("__info__", {})
        atoms.info.update(extra_info)

        # ------------------------------------------------------------------ #
        # 7.  Respect any user-defined r_data_keys renamings                 #
        # ------------------------------------------------------------------ #
        a2g_args = self.config.get("a2g_args", {}) or {}
        r_data_keys = a2g_args.get("r_data_keys", {})
        for custom, standard in r_data_keys.items():
            if standard in atoms.info:
                atoms.info[custom] = atoms.info[standard]
            elif standard in atoms.arrays:
                atoms.arrays[custom] = atoms.arrays[standard]

        return atoms

    def get_row_from_db(self, db_idx, el_idx):
        """Get a row from the database at the given indices."""
        db = self.dbs[db_idx]
        row_id = self.db_ids[db_idx][el_idx]
        if isinstance(db, LMDBDatabase):
            return db._get_row(row_id)  # pylint: disable=protected-access
        return db.get(row_id)

    @staticmethod
    def connect_db(
        address: str | Path, connect_args: dict | None = None
    ) -> ase.db.core.Database:
        if connect_args is None:
            connect_args = {}
        db_type = connect_args.get("type", "extract_from_name")
        if db_type in ("lmdb", "aselmdb") or (
            db_type == "extract_from_name"
            and str(address).rsplit(".", maxsplit=1)[-1] in ("lmdb", "aselmdb")
        ):
            return LMDBDatabase(address, readonly=True, **connect_args)

        return ase.db.connect(address, **connect_args)

    def __del__(self):
        for db in self.dbs:
            if hasattr(db, "close"):
                db.close()

    def sample_property_metadata(
        self,
    ) -> dict:  # Removed unused argument num_samples (W0613)
        """
        Sample property metadata from the database.

        This method was previously using the copy module which is now removed.
        """
        logging.warning(
            "You specified a folder of ASE dbs, so it's impossible to know which metadata to use. Using the first!"
        )
        if self.dbs[0].metadata == {}:
            return {}

        # Fixed unnecessary comprehension (R1721)
        return dict(self.dbs[0].metadata.items())



from matplotlib import pyplot as plt
from matplotlib.patches import FancyArrowPatch
from matplotlib import lines
from mpl_toolkits.mplot3d import proj3d
import numpy as np


class Arrow3D(FancyArrowPatch):
    def __init__(self, xs, ys, zs, *args, **kwargs):
        FancyArrowPatch.__init__(self, (0.0, 0.0), (0.0, 0.0), *args, **kwargs)
        self._verts3d = xs, ys, zs

    def draw(self, renderer):
        xs3d, ys3d, zs3d = self._verts3d
        xs, ys, zs = proj3d.proj_transform(xs3d, ys3d, zs3d, renderer.M)
        self.set_positions((xs[0], ys[0]), (xs[1], ys[1]))
        FancyArrowPatch.draw(self, renderer)


def plot_phonon_modes(structure, eigenvectors, q_vector,
                      supercell=(1, 1, 1),
                      draw_primitive=False,
                      vectors_scale=10,
                      by_element=True):

    atom_type = structure.get_atom_type_index(supercell=supercell)
    positions = structure.get_positions(supercell=supercell)
    masses = structure.get_masses(supercell=supercell)
    elements = structure.get_atomic_elements(supercell=supercell)
    np.set_printoptions(precision=8, suppress=True)

    cell_t = structure.get_cell()
    if draw_primitive:
        cell_t = structure.get_primitive_cell()

    for i_phonon in range(eigenvectors.shape[0]):

        fig = plt.figure(i_phonon+1)
        if draw_primitive:
            fig.suptitle('Primitive cell')
        else:
            fig.suptitle('Unit cell')

        ax = fig.add_subplot(111, projection='3d')

        color_atom=['g','b','m', 'c', 'y', 'k', 'w', 'g', 'b', 'm', 'c', 'y', 'k', 'w','g','b','m', 'c', 'y',
                    'k', 'w', 'g', 'b', 'm', 'c', 'y', 'k', 'w','g','b','m', 'c', 'y', 'k', 'w', 'g', 'b']

        if by_element:
            elements_unique = np.unique(elements, return_inverse=True)[1]
        else:
            elements_unique = atom_type

        # Atom positions
        for i, atom in enumerate(positions):
            ax.plot(atom[0][None], atom[1][None], atom[2][None], 'o', markersize=atom_radius[elements[i]]*30, color=color_atom[elements_unique[i]], alpha=0.8)

        # Cell frame
        for i in range(3):
            cell_side = [(0, cell_t[i, 0]), (0, cell_t[i, 1]), (0, cell_t[i, 2])]
            ax.plot3D(*cell_side, color='b')
            for j in range(3):
                if i != j:
                    cell_side = [(cell_t[i, l],
                                  cell_t[i, l]+cell_t[j, l]) for l in range(3)]

                    ax.plot3D(*cell_side, color='b')
                    for k in range(3):
                        if k != i and k != j and j > i:
                            cell_side = [(cell_t[i, l]+cell_t[j, l],
                                          cell_t[i, l]+cell_t[j, l]+cell_t[k, l]) for l in range(3)]

                            ax.plot3D(*cell_side, color='b')

        # Atom positions
        for i, position in enumerate(positions):
            eigenvector_atom = np.array(eigenvectors[i_phonon, atom_type[i], :])
            phase = 1.j * np.dot(position, q_vector)
            vector = (eigenvector_atom / np.sqrt(masses[atom_type[i]]) * np.exp(phase) * vectors_scale).real
 #           vector = np.dot(vector, np.linalg.inv(structure.get_primitive_matrix().T))
            a = Arrow3D([position[0], position[0]+vector[0]], [position[1], position[1]+vector[1]],
                        [position[2], position[2]+vector[2]], mutation_scale=20, lw=3, arrowstyle="-|>", color="r")
            ax.add_artist(a)

        # Legend
        atom_type_index_unique = np.unique(atom_type, return_index=True)[0]



        if by_element:
            atomic_types_unique = np.unique(elements, return_inverse=True)[0]
        else:
            atomic_types_unique = [elements[i] for i in atom_type_index_unique]

        legend_atoms =  [ lines.Line2D([0],[0], linestyle='none', c=color_atom[i], marker='o') for i, element in enumerate(atomic_types_unique)]
        ax.legend(legend_atoms, atomic_types_unique, numpoints = 1)

        # ax.set_axis_off()
        ax.set_xlabel('X')
        ax.set_ylabel('Y')
        ax.set_zlabel('Z')

        plt.title('Phonon {0}'.format(i_phonon+1))
        plt.axis('auto')
    plt.show()

    return


atom_radius = {
    'H':0.32,
    'He':0.93,
    'Li':1.23,
    'Be':0.9,
    'B':0.82,
    'C':0.77,
    'N':0.75,
    'O':0.73,
    'F':0.72,
    'Ne':0.71,
    'Na':1.54,
    'Mg':1.36,
    'Al':1.18,
    'Si':1.11,
    'P':1.06,
    'S':1.02,
    'Cl':0.99,
    'Ar':0.98,
    'K':2.03,
    'Ca':1.74,
    'Sc':1.44,
    'Ti':1.32,
    'V':1.22,
    'Cr':1.18,
    'Mn':1.17,
    'Fe':1.17,
    'Co':1.16,
    'Ni':1.15,
    'Cu':1.17,
    'Zn':1.25,
    'Ga':1.26,
    'Ge':1.22,
    'As':1.2,
    'Se':1.16,
    'Br':1.14,
    'Kr':1.12,
    'Rb':2.16,
    'Sr':1.91,
    'Y':1.62,
    'Zr':1.45,
    'Nb':1.34,
    'Mo':1.3,
    'Tc':1.27,
    'Ru':1.25,
    'Rh':1.25,
    'Pd':1.28,
    'Ag':1.34,
    'Cd':1.48,
    'In':1.44,
    'Sn':1.41,
    'Sb':1.4,
    'Te':1.36,
    'I':1.33,
    'Xe':1.31,
    'Cs':2.35,
    'Ba':1.98,
    'La':1.69,
    'Ce':1.65,
    'Pr':1.65,
    'Nd':1.64,
    'Pm':1.63,
    'Sm':1.62,
    'Eu':1.85,
    'Gd':1.61,
    'Tb':1.59,
    'Dy':1.59,
    'Ho':1.58,
    'Er':1.57,
    'Tm':1.56,
    'Yb':1.74,
    'Lu':1.56,
    'Hf':1.44,
    'Ta':1.34,
    'W':1.3,
    'Re':1.28,
    'Os':1.26,
    'Ir':1.27,
    'Pt':1.3,
    'Au':1.34,
    'Hg':1.49,
    'Tl':1.48,
    'Pb':1.47,
    'Bi':1.46,
    'Po':1.46,
    'At':1.45,
    'Rn':0.0,
    'Fr':0.0,
    'Ra':0.0,
    'Ac':0.0,
    'Th':1.65,
    'Pa':0.0,
    'U':1.42,
    'Np':0.0,
    'Pu':0.0,
    'Am':0.0,
    'Cm':0.0,
    'Bk':0.0,
    'Cf':0.0,
    'Es':0.0,
    'Fm':0.0,
    'Md':0.0,
    'No':0.0,
    'Lr':0.0,
    'Rf':0.0,
    'Db':0.0,
    'Sg':0.0,
    'Bh':0.0,
    'Hs':0.0,
    'Mt':0.0,
}


import os
import torch
import numpy as np
import tqdm
import matplotlib.pyplot as plt
from ase.io import read
from ase.units import GPa
from macer.calculator.factory import get_calculator

def calc_r2(true, pred):
    """
    Calculates the square of the Pearson correlation coefficient (r^2).
    This is invariant to constant shifts and scales, reflecting the true 'correlation'.
    """
    if len(true) < 2:
        return np.nan
    # Use correlation matrix [0,1] element
    c = np.corrcoef(true, pred)[0, 1]
    return c**2

def evaluate_model(data_path, ff="mattersim", model_path=None, device="cpu", output_dir="."):
    """
    General model evaluation logic.
    Calculates MAE and Correlation (r^2) for Energy, Forces, and Stress.
    Generates parity plots.
    """
    print(f"Loading data for evaluation: {data_path}")
    # Explicitly set format for XML to avoid ASE auto-detect confusion
    fmt = "vasp-xml" if data_path.lower().endswith(".xml") else None
    atoms_test = read(data_path, index=':', format=fmt)
    print(f"Total structures: {len(atoms_test)}")

    print(f"Initializing calculator: FF={ff}, Model={model_path if model_path else 'Default'}")
    
    # Dynamically build kwargs to handle different calculator requirements
    calc_kwargs = {"device": device}
    
    if ff == "mace":
        calc_kwargs["model_paths"] = [model_path] # For MACE compatibility
    else:
        calc_kwargs["model_path"] = model_path
        
    calc = get_calculator(ff_name=ff, **calc_kwargs)
    
    e_true_list, e_pred_list = [], []
    f_true_list, f_pred_list = [], []
    s_true_list, s_pred_list = [], []

    print(f"\nRunning inference via macer.calculator ({device})...")
    for atoms in tqdm.tqdm(atoms_test):
        try:
            e_true = atoms.get_potential_energy() / len(atoms) 
            f_true = atoms.get_forces()
        except Exception as e:
            print(f"Warning: Skipping structure due to missing DFT data: {e}")
            continue
            
        try:
            s_true = atoms.get_stress(voigt=False) / GPa
        except:
            s_true = None

        atoms.set_calculator(calc)
        e_pred = atoms.get_potential_energy() / len(atoms)
        f_pred = atoms.get_forces()
        try:
            s_pred = atoms.get_stress(voigt=False) / GPa
        except:
            s_pred = None

        e_true_list.append(e_true)
        e_pred_list.append(e_pred)
        f_true_list.extend(f_true.flatten())
        f_pred_list.extend(f_pred.flatten())
        if s_true is not None and s_pred is not None:
            s_true_list.extend(s_true.flatten())
            s_pred_list.extend(s_pred.flatten())

    if not e_true_list:
        print("Error: No valid predictions were made.")
        return None

    e_true_arr, e_pred_arr = np.array(e_true_list), np.array(e_pred_list)
    f_true_arr, f_pred_arr = np.array(f_true_list), np.array(f_pred_list)
    s_true_arr, s_pred_arr = np.array(s_true_list), np.array(s_pred_list)

    # Metrics (Using r^2 for correlation as requested)
    mae_e_raw = np.mean(np.abs(e_true_arr - e_pred_arr))
    r2_e = calc_r2(e_true_arr, e_pred_arr)
    
    shift = np.mean(e_true_arr) - np.mean(e_pred_arr)
    e_pred_shifted = e_pred_arr + shift
    mae_e_trend = np.mean(np.abs(e_true_arr - e_pred_shifted))

    mae_f = np.mean(np.abs(f_true_arr - f_pred_arr))
    r2_f = calc_r2(f_true_arr, f_pred_arr)
    
    mae_s, r2_s = None, None
    if s_true_list:
        mae_s = np.mean(np.abs(s_true_arr - s_pred_arr))
        r2_s = calc_r2(s_true_arr, s_pred_arr)

    # Plotting helper
    def _write_parity_dat(true, pred, filename):
        base, _ = os.path.splitext(filename)
        dat_path = os.path.join(output_dir, f"{base}.dat")
        with open(dat_path, "w") as f:
            f.write("# Input Prediction\n")
            for t, p in zip(true, pred):
                f.write(f"{t:.8f} {p:.8f}\n")

    def plot_parity(true, pred, title, xlabel, ylabel, filename, mae, r2):
        plt.figure(figsize=(6, 6))
        ax = plt.gca()
        ax.scatter(true, pred, alpha=0.5, s=10, edgecolors='none')
        combined = np.concatenate([true, pred])
        lims = [np.min(combined), np.max(combined)]
        ax.plot(lims, lims, 'k--', alpha=0.75, zorder=0)
        ax.set_aspect('equal', adjustable='box')
        ax.set_title(f"{title}\nMAE: {mae:.4f} | r²: {r2:.4f}")
        ax.set_xlabel(xlabel)
        ax.set_ylabel(ylabel)
        plt.tight_layout()
        plt.savefig(os.path.join(output_dir, filename))
        plt.close()
        _write_parity_dat(true, pred, filename)

    print(f"Generating enhanced parity plots in {os.path.abspath(output_dir)}...")
    plot_parity(e_true_arr, e_pred_shifted, "Absolute Energy (Shifted)", "Input", "Prediction", "eval_parity_energy_abs.pdf", mae_e_trend, r2_e)
    e_true_rel = e_true_arr - np.min(e_true_arr)
    e_pred_rel = e_pred_arr - np.min(e_pred_arr)
    plot_parity(e_true_rel, e_pred_rel, "Relative Energy (min=0)", "Input", "Prediction", "eval_parity_energy_rel.pdf", mae_e_trend, r2_e)
    plot_parity(f_true_arr, f_pred_arr, "Atomic Forces", "Input", "Prediction", "eval_parity_forces.pdf", mae_f, r2_f)
    if s_true_list:
        plot_parity(s_true_arr, s_pred_arr, "Virial Stress", "Input", "Prediction", "eval_parity_stress.pdf", mae_s, r2_s)

    results = []
    results.append("-" * 55)
    results.append(f"{ 'Metric':<15} | {'MAE':<12} | {'r² (Corr)':<12}")
    results.append("-" * 55)
    results.append(f"{ 'Energy (Raw)':<15} | {mae_e_raw:12.6f} | {r2_e:12.4f}")
    results.append(f"{ 'Energy (Trend)':<15} | {mae_e_trend:12.6f} | {r2_e:12.4f}")
    results.append(f"{ 'Forces':<15} | {mae_f:12.6f} | {r2_f:12.4f}")
    if mae_s is not None:
        results.append(f"{ 'Stress':<15} | {mae_s:12.6f} | {r2_s:12.4f}")
    else:
        results.append(f"{ 'Stress':<15} | Not available in data")
    results.append("-" * 55)
    
    return "\n".join(results)



from typing import Optional

import torch

from macer.externals.e3nn_legacy import o3


class CartesianTensor(o3.Irreps):
    r"""representation of a cartesian tensor into irreps

    Parameters
    ----------
    formula : str

    Examples
    --------

    >>> import torch
    >>> CartesianTensor("ij=-ji")
    1x1e

    >>> x = CartesianTensor("ijk=-jik=-ikj")
    >>> x.from_cartesian(torch.ones(3, 3, 3))
    tensor([0.])

    >>> x.from_vectors(torch.ones(3), torch.ones(3), torch.ones(3))
    tensor([0.])

    >>> x = CartesianTensor("ij=ji")
    >>> t = torch.arange(9).to(torch.float).view(3,3)
    >>> y = x.from_cartesian(t)
    >>> z = x.to_cartesian(y)
    >>> torch.allclose(z, (t + t.T)/2, atol=1e-5)
    True
    """
    # pylint: disable=abstract-method

    # These are set in __new__
    formula: str
    indices: str

    def __new__(
        # pylint: disable=signature-differs
        cls,
        formula,
    ):
        indices = formula.split("=")[0].replace("-", "")
        rtp = o3.ReducedTensorProducts(formula, **{i: "1o" for i in indices})
        ret = super().__new__(cls, rtp.irreps_out)
        ret.formula = formula
        ret.indices = indices
        return ret

    def from_cartesian(self, data, rtp=None):
        r"""convert cartesian tensor into irreps

        Parameters
        ----------
        data : `torch.Tensor`
            cartesian tensor of shape ``(..., 3, 3, 3, ...)``

        Returns
        -------
        `torch.Tensor`
            irreps tensor of shape ``(..., self.dim)``
        """
        if rtp is None:
            rtp = self.reduced_tensor_products(data)

        Q = rtp.change_of_basis.flatten(-len(self.indices))
        return data.flatten(-len(self.indices)) @ Q.T

    def from_vectors(self, *xs, rtp=None):
        r"""convert :math:`x_1 \otimes x_2 \otimes x_3 \otimes \dots`

        Parameters
        ----------
        xs : list of `torch.Tensor`
            list of vectors of shape ``(..., 3)``

        Returns
        -------
        `torch.Tensor`
            irreps tensor of shape ``(..., self.dim)``
        """
        if rtp is None:
            rtp = self.reduced_tensor_products(xs[0])

        return rtp(*xs)  # pylint: disable=not-callable

    def to_cartesian(self, data, rtp=None):
        r"""convert irreps tensor to cartesian tensor

        This is the symmetry-aware inverse operation of ``from_cartesian()``.

        Parameters
        ----------
        data : `torch.Tensor`
            irreps tensor of shape ``(..., D)``, where D is the dimension of the irreps,
            i.e. ``D=self.dim``.

        Returns
        -------
        `torch.Tensor`
            cartesian tensor of shape ``(..., 3, 3, 3, ...)``
        """
        if rtp is None:
            rtp = self.reduced_tensor_products(data)

        Q = rtp.change_of_basis
        cartesian_tensor = data @ Q.flatten(-len(self.indices))

        shape = list(data.shape[:-1]) + list(Q.shape[1:])
        cartesian_tensor = cartesian_tensor.view(shape)

        return cartesian_tensor

    def reduced_tensor_products(self, data: Optional[torch.Tensor] = None) -> o3.ReducedTensorProducts:
        r"""reduced tensor products

        Returns
        -------
        `e3nn.ReducedTensorProducts`
            reduced tensor products
        """
        rtp = o3.ReducedTensorProducts(self.formula, **{i: "1o" for i in self.indices})
        if data is not None:
            rtp = rtp.to(device=data.device, dtype=data.dtype)
        return rtp



import sys
import logging
from phonopy.structure.symmetry import Symmetry

def apply_dynaphopy_patch():
    """
    Patches the installed dynaphopy package at runtime to fix compatibility issues
    with Phonopy 2.x (specifically the removal of get_reciprocal_operations).
    """
    try:
        # Try bundled version first
        import macer.externals.dynaphopy_bundled.interface.phonopy_link as ph_link
    except (ImportError, ModuleNotFoundError):
        try:
            # Fallback to system version
            import dynaphopy.interface.phonopy_link as ph_link
        except (ImportError, ModuleNotFoundError):
            # Dynaphopy not installed, nothing to patch
            return
        except Exception:
            return
    except Exception:
        # Any other error
        return

    # 2. Check if the patch is needed (look for the offending code logic or just force patch)
    # We redefine the specific function that causes the crash.
    
    # Original signature: get_equivalent_q_points_by_symmetry(q_point, bulk, symprec=1e-5)
    def patched_get_equivalent_q_points_by_symmetry(q_point, bulk, symprec=1e-5):
        # --- PATCH START ---
        import numpy as np
        
        # Ensure q_point is a numpy array (NumPy 2.x safety)
        q_point = np.array(q_point, dtype=float)
        
        # DynaPhoPy passes its own 'Structure' object, but Phonopy 2.x's Symmetry class
        # expects a PhonopyAtoms-like object with attributes like 'magnetic_moments'.
        # We must convert 'bulk' to PhonopyAtoms to avoid AttributeError.
        
        from phonopy.structure.atoms import PhonopyAtoms
        
        # Convert DynaPhoPy Structure -> PhonopyAtoms
        # DynaPhoPy Structure attributes: get_cell(), get_scaled_positions(), get_atomic_numbers()
        phonopy_cell = PhonopyAtoms(cell=bulk.get_cell(),
                                    scaled_positions=bulk.get_scaled_positions(),
                                    numbers=bulk.get_atomic_numbers())
        
        # Use property .reciprocal_operations instead of getter
        try:
            # Phonopy 2.x style
            ops = Symmetry(phonopy_cell, symprec=symprec).reciprocal_operations
        except AttributeError:
            # Fallback for very old Phonopy (unlikely, but safe)
            ops = Symmetry(phonopy_cell, symprec=symprec).get_reciprocal_operations()
        
        for operation_matrix in ops:
            q_point_rot =  np.dot(operation_matrix, q_point)
            diff = q_point_rot - q_point
            if (np.abs(diff - np.rint(diff)) < symprec).all():
                return [q_point]  # Return as a list containing the vector
        return [q_point]          # Return as a list containing the vector
        # --- PATCH END ---


    # 3. Apply the Monkey Patch
    # We need numpy for the function above to work, ensuring it's in the module's scope
    import numpy as np
    
    # Inject numpy into the module scope if it's missing (usually it's there)
    ph_link.np = np 
    
    # Swap the broken function with our fixed one
    ph_link.get_equivalent_q_points_by_symmetry = patched_get_equivalent_q_points_by_symmetry
    
    # logger = logging.getLogger("macer")
    # logger.debug("Applied runtime patch to dynaphopy.interface.phonopy_link for Phonopy 2.x compatibility.")
    # print("DEBUG: Applied runtime patch to dynaphopy.interface.phonopy_link")




"""Utility functions for 2nd order force constants."""

from __future__ import annotations

import itertools

import numpy as np
from numpy.typing import NDArray
from scipy.sparse import csr_array, kron

from macer.externals.symfc_bundled.spg_reps import SpgRepsO2
from macer.externals.symfc_bundled.utils.cutoff_tools import FCCutoff

from .utils import get_indep_atoms_by_lat_trans


def get_lat_trans_decompr_indices(trans_perms: NDArray) -> NDArray:
    """Return indices to de-compress compressed matrix by lat-trans-sym.

    Usage
    -----
    vec[indices] of shape (n_a*N*9,) gives an array of shape=(N**2*9,).
    1/sqrt(n_lp) must be multiplied manually after decompression to mimic
    get_lat_trans_compr_matrix.

    Parameters
    ----------
    trans_perms : ndarray
        Permutation of atomic indices by lattice translational symmetry.
        dtype='intc'.
        shape=(n_l, N), where n_l and N are the numbers of lattice points and
        atoms in supercell.

    Returns
    -------
    indices : ndarray
        Indices of n_a * N9 elements.
        shape=(N^2*9,), dtype='int_'.

    """
    indep_atoms = get_indep_atoms_by_lat_trans(trans_perms)
    n_a = len(indep_atoms)
    N = trans_perms.shape[1]
    n_lp = N // n_a
    size_row = (N * 3) ** 2

    n = 0
    indices = np.zeros(size_row, dtype="int_")
    for i_patom in indep_atoms:
        index_shift_i = trans_perms[:, i_patom] * N * 9
        for j in range(N):
            index_shift = index_shift_i + trans_perms[:, j] * 9
            for ab in range(9):
                indices[index_shift + ab] = n
                n += 1
    assert n * n_lp == size_row
    return indices


def get_lat_trans_compr_indices(trans_perms: NDArray) -> NDArray:
    """Return indices to compress matrix by lat-trans-sym.

    Usage
    -----
    vec[indices] of shape (N**2*9,) vec gives an array of shape=(n_a*N*9, n_lp).
    1/sqrt(n_lp) must be multiplied manually after compression to mimic
    get_lat_trans_compr_matrix.

    Parameters
    ----------
    trans_perms : ndarray
        Permutation of atomic indices by lattice translational symmetry.
        dtype='intc'. shape=(n_l, N), where n_l and N are the numbers of lattice
        points and atoms in supercell.

    Returns
    -------
    indices : ndarray
        shape=(n_a*N9, n_lp), dtype='int_'.

    """
    indep_atoms = get_indep_atoms_by_lat_trans(trans_perms)
    n_a = len(indep_atoms)
    N = trans_perms.shape[1]
    n_lp = N // n_a
    size_row = (N * 3) ** 2

    n = 0
    indices = np.zeros((n_a * N * 9, n_lp), dtype="int_")
    for i_patom in indep_atoms:
        for j in range(N):
            for ab in range(9):
                indices[n, :] = (
                    trans_perms[:, i_patom] * 9 * N + trans_perms[:, j] * 9 + ab
                )
                n += 1
    assert n * n_lp == size_row
    return indices


def get_lat_trans_compr_matrix(decompr_idx: NDArray, N: int, n_lp: int) -> csr_array:
    """Return compression matrix by lattice translation symmetry.

    `decompr_idx` is obtained by `get_lat_trans_decompr_indices`.

    Matrix shape is (NN33, n_a*N33), where n_a is the number of independent
    atoms by lattice translation symmetry.

    Data order is (N, N, 3, 3, n_a, N, 3, 3) if it is in dense array.

    """
    NN9 = N**2 * 9
    compression_mat = csr_array(
        (
            np.full(NN9, 1 / np.sqrt(n_lp), dtype="double"),
            (np.arange(NN9, dtype=int), decompr_idx),
        ),
        shape=(NN9, NN9 // n_lp),
        dtype="double",
    )
    return compression_mat


def get_lat_trans_compr_matrix_O2(trans_perms: NDArray):
    """Return lat trans compression matrix."""
    n_lp, N = trans_perms.shape
    decompr_idx = get_lat_trans_decompr_indices(trans_perms)
    c_trans = get_lat_trans_compr_matrix(decompr_idx, N, n_lp)
    return c_trans


def get_perm_compr_matrix(natom: int) -> csr_array:
    """Return compression matrix by permutation symmetry.

    Parameters
    ----------
    natom : int
        Number of atoms in supercell.

    Matrix shape is (NN33,(N*3)(N*3+1)/2).
    Non-zero only ijab and jiba column elements for ijab rows.
    Rows upper right NN33 matrix elements are selected for rows.

    For the computational performance, get_perm_compr_matrix is implemented in a
    tricky way effectively using numpy features, and so it is not easy to read.
    What is expected may be found reading _get_perm_compr_matrix_reference.

    """
    N = natom
    A = np.transpose(
        np.arange(N**2 * 9).reshape(N, N, 3, 3), axes=(0, 2, 1, 3)
    ).reshape(N * 3, N * 3)
    ut = np.triu_indices_from(A, k=1)
    diag = np.diagonal(A)
    row = np.hstack((np.stack((A[ut], A.T[ut]), axis=1).ravel(), diag))
    col = np.hstack(
        (
            np.repeat(np.arange(len(ut[0]), dtype=int), 2),
            np.arange(len(ut[0]), len(ut[0]) + len(diag), dtype=int),
        )
    )
    data = np.hstack((np.full(len(ut[0]) * 2, np.sqrt(2) / 2), np.full(len(diag), 1)))
    return csr_array(
        (data, (row, col)),
        shape=(N**2 * 9, (N * 3 * (N * 3 + 1)) // 2),
        dtype="double",
    )


def _get_atomic_lat_trans_decompr_indices(trans_perms: NDArray) -> NDArray:
    """Return indices to de-compress compressed matrix by atom-lat-trans-sym.

    This is atomic permutation only version of get_lat_trans_decompr_indices.

    Usage
    -----
    vec[indices] of shape (n_a*N,) gives an array of shape=(N**2,).
    1/sqrt(n_lp) must be multiplied manually after decompression.

    Parameters
    ----------
    trans_perms : ndarray
        Permutation of atomic indices by lattice translational symmetry.
        dtype='intc'.
        shape=(n_l, N), where n_l and N are the numbers of lattice points and
        atoms in supercell.

    Returns
    -------
    indices : ndarray
        Indices of n_a * N elements.
        shape=(N^2*,), dtype='int_'.

    """
    indep_atoms = get_indep_atoms_by_lat_trans(trans_perms)
    n_lp, N = trans_perms.shape
    size_row = N**2

    n = 0
    indices = np.zeros(size_row, dtype="int_")
    for i_patom in indep_atoms:
        index_shift_i = trans_perms[:, i_patom] * N
        for j in range(N):
            index_shift = index_shift_i + trans_perms[:, j]
            indices[index_shift] = n
            n += 1
    assert n * n_lp == size_row
    return indices


def get_compr_coset_reps_sum(spg_reps: SpgRepsO2) -> csr_array:
    """Return compressed projector of coset reps sum."""
    trans_perms = spg_reps.translation_permutations
    n_lp, N = trans_perms.shape
    size = N**2 * 9 // n_lp
    coset_reps_sum = csr_array(([], ([], [])), shape=(size, size), dtype="double")
    atomic_decompr_idx = _get_atomic_lat_trans_decompr_indices(trans_perms)
    C = csr_array(
        (
            np.ones(N**2, dtype=int),
            (np.arange(N**2, dtype=int), atomic_decompr_idx),
        ),
        shape=(N**2, N**2 // n_lp),
    )
    factor = 1 / n_lp / len(spg_reps.unique_rotation_indices)
    for i, _ in enumerate(spg_reps.unique_rotation_indices):
        mat = spg_reps.get_sigma2_rep(i)
        mat = mat @ C
        mat = C.T @ mat
        coset_reps_sum += kron(mat, spg_reps.r_reps[i] * factor)

    return coset_reps_sum


def _get_perm_compr_matrix_reference(natom: int) -> csr_array:
    """Return compression matrix by permutation symmetry.

    This is a reference implementation of get_perm_compr_matrix. The order of
    columns is difference from get_perm_compr_matrix, but it is OK if C.T@C
    is the same.

    Matrix shape is (NN33,(N*3)(N*3+1)/2). Non-zero only ijab and jiba column
    elements for ijab rows. Rows upper right NN33 matrix elements are selected
    for rows.

    """

    def to_serial(i: int, a: int, j: int, b: int, natom: int) -> int:
        """Return NN33-1D index."""
        return (i * 9 * natom) + (j * 9) + (a * 3) + b

    col, row, data = [], [], []
    val = np.sqrt(2) / 2
    size_row = natom**2 * 9

    n = 0
    for ia, jb in itertools.combinations_with_replacement(range(natom * 3), 2):
        i_i = ia // 3
        i_a = ia % 3
        i_j = jb // 3
        i_b = jb % 3
        col.append(n)
        row.append(to_serial(i_i, i_a, i_j, i_b, natom))
        if i_i == i_j and i_a == i_b:
            data.append(1)
        else:
            data.append(val)
            col.append(n)
            row.append(to_serial(i_j, i_b, i_i, i_a, natom))
            data.append(val)
        n += 1
    if (natom * 3) % 2 == 1:
        assert (natom * 3) * ((natom * 3 + 1) // 2) == n, f"{natom}, {n}"
    else:
        assert ((natom * 3) // 2) * (natom * 3 + 1) == n, f"{natom}, {n}"
    return csr_array((data, (row, col)), shape=(size_row, n), dtype="double")


def get_compr_coset_projector_O2(
    spg_reps: SpgRepsO2,
    atomic_decompr_idx: NDArray | None = None,
    fc_cutoff: FCCutoff | None = None,
    c_pt: csr_array | None = None,
) -> csr_array:
    """Return compr matrix of sum of coset reps."""
    trans_perms = spg_reps.translation_permutations
    n_lp, N = trans_perms.shape
    size = N**2 * 9 // n_lp if c_pt is None else c_pt.shape[1]  # type: ignore
    coset_reps_sum = csr_array((size, size), dtype="double")

    if atomic_decompr_idx is None:
        _atomic_decompr_idx = _get_atomic_lat_trans_decompr_indices(trans_perms)
    else:
        _atomic_decompr_idx = atomic_decompr_idx

    if fc_cutoff is None:
        nonzero = None
        size_data = N**2
        col = _atomic_decompr_idx
    else:
        nonzero = fc_cutoff.nonzero_atomic_indices_fc2()
        size_data = np.count_nonzero(nonzero)
        col = _atomic_decompr_idx[nonzero]

    factor = 1 / n_lp / len(spg_reps.unique_rotation_indices)
    for i, _ in enumerate(spg_reps.unique_rotation_indices):
        permutation = spg_reps.get_sigma2_rep(i, nonzero=nonzero)
        mat = csr_array(
            (
                np.ones(size_data, dtype="int_"),
                (_atomic_decompr_idx[permutation], col),  # type: ignore
            ),
            shape=(N**2 // n_lp, N**2 // n_lp),
            dtype="int_",
        )
        mat = kron(mat, spg_reps.r_reps[i] * factor)
        if c_pt is not None:
            mat = c_pt.T @ mat @ c_pt

        coset_reps_sum += mat

    return coset_reps_sum



"""
Macer: Machine-learning accelerated Atomic Computational Environment for automated Research workflows
Copyright (c) 2025 The Macer Package Authors
Author: Soungmin Bae <soungminbae@gmail.com>
License: MIT
"""

import sys
import os
import shutil
from argparse import Namespace
from pathlib import Path
import yaml

from pymatgen.core import IStructure, Structure
from monty.serialization import loadfn

# pydefect imports
from pydefect.cli.main_functions import (
    make_supercell,
    make_defect_energy_infos_main_func,
    make_defect_energy_summary_main_func
)
from pydefect.cli.vasp.main_vasp_functions import make_defect_entries, make_calc_results
from pydefect.input_maker.defect_set_maker import DefectSetMaker
from pydefect.chem_pot_diag.chem_pot_diag import StandardEnergies
from pydefect.analyzer.unitcell import Unitcell
from pydefect.analyzer.band_edge_states import PerfectBandEdgeState, EdgeInfo, OrbitalInfo

# macer imports
from macer.pydefect.utils import run_macer_relax, write_summary_at_vertices, get_unique_dir_name, dim_to_matrix, make_manual_supercell_info
from macer.utils.validation import check_poscar_format

def run_defect_workflow(args):
    """
    Auto Defect workflow.
    args requires: poscar, std_energies, target_vertices, matrix, min_atoms, max_atoms, analyze_symmetry, sites_yaml_filename, doping
    """
    
    if getattr(args, "dim", None):
        if args.matrix:
            raise ValueError("Use either --matrix or --dim, not both.")

    from macer.utils.struct_tools import resolve_structure_inputs
    resolved = resolve_structure_inputs(poscar_paths=[args.poscar])
    if not resolved:
        raise FileNotFoundError("Error: No valid POSCAR provided.")
    args.poscar = resolved[0]
    input_poscar_path = Path(args.poscar).absolute()
    if not input_poscar_path.exists():
        raise FileNotFoundError(f"Error: {input_poscar_path} not found.")
        
    # Validate POSCAR format (VASP 5 check)
    try:
        check_poscar_format(input_poscar_path)
    except ValueError as e:
        raise ValueError(f"\nError: {e}\n")
    
    std_energies_path = Path(args.std_energies).absolute() if args.std_energies else None
    target_vertices_path = Path(args.target_vertices).absolute() if args.target_vertices else None

    sites_yaml_path = None
    if args.sites_yaml_filename:
        sites_yaml_path = Path(args.sites_yaml_filename).absolute()
        if not sites_yaml_path.exists():
            raise FileNotFoundError(f"Error: {sites_yaml_path} not found.")

    # 1. Setup Directory
    try:
        structure = Structure.from_file(str(input_poscar_path))
        formula = structure.composition.reduced_formula
        input_poscar_name = input_poscar_path.name
        
        dopants_part = ""
        if args.doping:
            dopants_part = f"-DOPANG={'_'.join(args.doping)}"

        if getattr(args, "output_dir", None):
            base_dir_name = Path(args.output_dir)
        else:
            base_dir_name = input_poscar_path.parent / f"DEFECT-{input_poscar_name}-formula={formula}{dopants_part}-mlff={args.ff}"
            
        work_dir = get_unique_dir_name(base_dir_name)
        
        print(f"Creating workspace: {work_dir}")
        work_dir.mkdir()
        os.chdir(work_dir)
        
        # Copy input POSCAR to POSCAR-unitcell
        shutil.copy(str(input_poscar_path), "POSCAR-unitcell")
        
        # Update poscar_path to the local copy for subsequent steps
        poscar_path = Path("POSCAR-unitcell")
        dopants = args.doping if args.doping else []

    except Exception as e:
        raise RuntimeError(f"Initialization failed: {e}")

    print("--- Step 1: Make Supercell ---")
    unitcell = IStructure.from_file(str(poscar_path))
    if getattr(args, "dim", None):
        make_manual_supercell_info(
            poscar_path,
            dim=args.dim,
            output_dir=Path("."),
            analyze_symmetry=True
        )
    else:
        args_sc = Namespace(unitcell=unitcell, matrix=args.matrix, 
                            min_num_atoms=args.min_atoms, max_num_atoms=args.max_atoms, 
                            analyze_symmetry=args.analyze_symmetry, 
                            dopants=dopants,
                            sites_yaml_filename=str(sites_yaml_path) if sites_yaml_path else None)
        make_supercell(args_sc)
    
    if not Path("supercell_info.json").exists():
        raise RuntimeError("Error: supercell_info.json not created.")

    if getattr(args, "interstitial", False):
        from macer.pydefect.utils import run_voronoi_interstitial_finding
        run_voronoi_interstitial_finding(Path("supercell_info.json"))

    print("\n--- Step 2: Make Defect Set (Charge 0 only) ---")
    supercell_info = loadfn("supercell_info.json")
    
    # 1. Generate vacancies (using a copy without interstitials to be safe)
    from dataclasses import replace
    si_no_interstitial = replace(supercell_info, interstitials=[])
    maker_vacancies = DefectSetMaker(si_no_interstitial)
    
    # 2. Generate others (interstitials and substitutions)
    host_elements = [str(e) for e in supercell_info.unitcell_structure.composition.elements]
    effective_dopants = list(set(host_elements + (dopants if dopants else [])))
    maker_others = DefectSetMaker(supercell_info, dopants=effective_dopants)
    
    # 3. Merge
    all_defects = {d.name: d for d in list(maker_vacancies.defect_set) + list(maker_others.defect_set)}
    defect_dict = {name: [0] for name in all_defects}
    
    with open("defect_in.yaml", "w") as f:
        yaml.dump(defect_dict, f)
    print(f"defect_in.yaml created with {len(defect_dict)} defects: {list(defect_dict.keys())}")

    print("\n--- Step 3: Make Defect Entries ---")
    make_defect_entries(Namespace())
    
    # Identify directories to calculate
    dirs_to_calc = []
    if Path("perfect").exists() and Path("perfect/POSCAR").exists():
        dirs_to_calc.append(Path("perfect"))
    
    for name in defect_dict:
        d = Path(f"{name}_0")
        if d.exists() and (d / "POSCAR").exists():
            dirs_to_calc.append(d)
    
    print(f"Directories to calculate: {[d.name for d in dirs_to_calc]}")

    print("\n--- Step 4: Run Macer Relax ---")
    
    # Use ISIF=2 for defects
    fmax_defect_val = getattr(args, "fmax_defect", None)
    if fmax_defect_val is None:
        fmax_defect_val = args.fmax

    successful_dirs = run_macer_relax(
        dirs_to_calc, 
        isif=2, 
        supercell_info=supercell_info, 
        verbose=True, 
        fmax=fmax_defect_val,
        ff=args.ff,
        model_path=args.model,
        device=args.device,
        modal=args.modal,
        n_jobs=getattr(args, "n_jobs", 1),
    )

    print("\n--- Step 4.5: Generate calc_results.json ---")
    if successful_dirs:
        make_calc_results(Namespace(dirs=successful_dirs, verbose=False, check_calc_results=False))
    else:
        print("No successful calculations found.")
        return

    if not (std_energies_path and std_energies_path.exists()):
        print(f"Standard energies not found. Exiting after calculations.")
        return

    print("\n--- Step 5: Generate Defect Energy Info ---")
    
    # Create unitcell.yaml with correct system name
    formula = unitcell.composition.reduced_formula
    system_name = f"{poscar_path.name}-{formula}"
    
    dummy_unitcell_str = f"""system: {system_name}
vbm: 0.0
cbm: 5.0
ele_dielectric_const: [[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]]
ion_dielectric_const: [[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]]
"""
    Path("unitcell.yaml").write_text(dummy_unitcell_str)
    
    perfect_calc_results_path = Path("perfect/calc_results.json")
    if not perfect_calc_results_path.exists():
        print("Error: perfect/calc_results.json not found.")
        return

    perfect_calc_results = loadfn(str(perfect_calc_results_path))
    unitcell_obj = Unitcell.from_yaml("unitcell.yaml")
    std_energies_obj = StandardEnergies.from_yaml(str(std_energies_path))
    
    defect_dirs = [d for d in successful_dirs if d.name != "perfect"]
    
    args_dei = Namespace(dirs=defect_dirs, check_calc_results=False, perfect_calc_results=perfect_calc_results,
                         std_energies=std_energies_obj, unitcell=unitcell_obj, verbose=False)
    
    make_defect_energy_infos_main_func(args_dei)
    print("Done. Defect energy infos generated.")

    print("\n--- Step 6: Generate Defect Energy Summary ---")
    
    if target_vertices_path and target_vertices_path.exists():
        # Dummy Band Edge State
        vbm_info = EdgeInfo(band_idx=0, kpt_coord=(0.0, 0.0, 0.0),
                            orbital_info=OrbitalInfo(energy=0.0, orbitals={}, occupation=1.0))
        cbm_info = EdgeInfo(band_idx=0, kpt_coord=(0.0, 0.0, 0.0),
                            orbital_info=OrbitalInfo(energy=5.0, orbitals={}, occupation=0.0))
        p_state = PerfectBandEdgeState(vbm_info=vbm_info, cbm_info=cbm_info)
        
        args_des = Namespace(
            dirs=defect_dirs,
            verbose=False,
            target_vertices_yaml=str(target_vertices_path),
            unitcell=unitcell_obj,
            p_state=p_state
        )
        
        make_defect_energy_summary_main_func(args_des)
        print("defect_energy_summary.json created.")
    else:
        print("Target vertices not found. Skipping standard Defect Energy Summary generation.")

    print("\n--- Step 7: Write summary ---")
    summary_path = Path("defect_energy_summary.json")
    if summary_path.exists():
        write_summary_at_vertices(summary_path)
    elif not (target_vertices_path and target_vertices_path.exists()):
        print("Using simplified summary generation (Standard Energies only)...")
        try:
            from macer.pydefect.utils import write_simple_defect_summary
            write_simple_defect_summary(Path("."))
        except ImportError:
            print("Error: write_simple_defect_summary not found in utils.")
    else:
        print("defect_energy_summary.json not found, skipping final summary generation.")



__version__ = '1.18.0'
import macer.externals.dynaphopy_bundled.projection as projection
import macer.externals.dynaphopy_bundled.parameters as parameters
import macer.externals.dynaphopy_bundled.interface.phonopy_link as pho_interface
import macer.externals.dynaphopy_bundled.interface.iofile as reading
import macer.externals.dynaphopy_bundled.analysis.energy as energy
import macer.externals.dynaphopy_bundled.analysis.fitting as fitting
import macer.externals.dynaphopy_bundled.analysis.modes as modes
import macer.externals.dynaphopy_bundled.analysis.coordinates as trajdist
import macer.externals.dynaphopy_bundled.analysis.thermal_properties as thm
import numpy as np
import matplotlib.pyplot as plt


from macer.externals.dynaphopy_bundled.power_spectrum import power_spectrum_functions
try:
    from scipy.integrate import simps
except ImportError:
    from scipy.integrate import simpson as simps


class Quasiparticle:
    def __init__(self,
                 dynamic,
                 last_steps=None,
                 vc=None):

        self._dynamic = dynamic
        self._vc = vc
        self._eigenvectors = None
        self._frequencies = None
        self._vq = None
        self._power_spectrum_phonon = None
        self._power_spectrum_wave_vector = None
        self._power_spectrum_direct = None
        self._power_spectrum_partials = None
        self._bands = None
        self._renormalized_bands = None
        self._renormalized_force_constants = None
        self._commensurate_points_data = None
        self._temperature = None
        self._force_constants_qha = None
        self._parameters = parameters.Parameters()
        self.crop_trajectory(last_steps)
        #  print('Using {0} time steps for calculation'.format(len(self.dynamic.velocity)))

    # Crop trajectory
    def crop_trajectory(self, last_steps):
        if self._vc is None:
            self._dynamic.crop_trajectory(last_steps)
            print("Using {0} steps".format(len(self._dynamic.velocity)))
        else:
            if last_steps is not None:
                self._vc = self._vc[-last_steps:, :, :]
            print("Using {0} steps".format(len(self._vc)))

    # Memory clear methods
    def full_clear(self):
        self._eigenvectors = None
        self._frequencies = None
        self._vc = None
        self._vq = None
        self._power_spectrum_direct = None
        self._power_spectrum_wave_vector = None
        self._power_spectrum_phonon = None

    def power_spectra_clear(self):
        self._power_spectrum_phonon = None
        self._power_spectrum_wave_vector = None
        self._power_spectrum_direct = None
        self.force_constants_clear()

    def force_constants_clear(self):
        self._renormalized_force_constants = None
        self._commensurate_points_data = None
        self.bands_clear()

    def bands_clear(self):
        self._bands = None
        self._renormalized_bands = None

    # Properties
    @property
    def dynamic(self):
        return self._dynamic

    @property
    def parameters(self):
        return self._parameters

    def set_NAC(self, NAC):
        """
        Turns on or off Non-Analitic corrections

        :param NAC: True or False
        :return: None
        """
        self._bands = None
        self.parameters.use_NAC = NAC

    def write_to_xfs_file(self, file_name):
        reading.write_xsf_file(file_name, self.dynamic.structure)

    def save_velocity_hdf5(self, file_name, save_trajectory=True):
        if save_trajectory:
            trajectory = self.dynamic.trajectory
        else:
            trajectory = None

        reading.save_data_hdf5(file_name,
                               self.dynamic.get_time(),
                               self.dynamic.get_supercell_matrix(),
                               velocity=self.dynamic.velocity,
                               trajectory=trajectory)

        print("Velocity saved in file " + file_name)

    def save_vc_hdf5(self, file_name):

        reading.save_data_hdf5(file_name,
                               self.dynamic.get_time(),
                               self.dynamic.get_supercell_matrix(),
                               vc=self.get_vc(),
                               reduced_q_vector=self.get_reduced_q_vector())

        print("Projected velocity saved in file " + file_name)

    def set_number_of_mem_coefficients(self, coefficients):
        self.power_spectra_clear()
        self.parameters.number_of_coefficients_mem = coefficients

    def set_projection_onto_atom_type(self, atom_type):
        if atom_type in range(self.dynamic.structure.get_number_of_primitive_atoms()):
            self.parameters.project_on_atom = atom_type
        else:
            print('Atom type {} does not exist'.format(atom_type))
            exit()

    def _set_frequency_range(self, frequency_range):
        if not np.array_equiv(np.array(frequency_range), np.array(self.parameters.frequency_range)):
            self.power_spectra_clear()
            self.parameters.frequency_range = frequency_range

    def set_spectra_resolution(self, resolution):
        limits = [self.get_frequency_range()[0], self.get_frequency_range()[-1]]
        self.parameters.spectrum_resolution = resolution
        self._set_frequency_range(np.arange(limits[0], limits[1] + resolution, resolution))

    def set_frequency_limits(self, limits):
        resolution = self.parameters.spectrum_resolution
        self._set_frequency_range(np.arange(limits[0], limits[1] + resolution, resolution))

    def get_frequency_range(self):
        return self.parameters.frequency_range

    # Wave vector related methods
    def set_reduced_q_vector(self, q_vector):
        if len(q_vector) == len(self.parameters.reduced_q_vector):
            if (np.array(q_vector) != self.parameters.reduced_q_vector).any():
                self.full_clear()

        self.parameters.reduced_q_vector = np.array(q_vector)

    def get_reduced_q_vector(self):
        return self.parameters.reduced_q_vector

    def get_q_vector(self):
        return np.dot(self.parameters.reduced_q_vector,
                      2.0 * np.pi * np.linalg.inv(self.dynamic.structure.get_primitive_cell()).T)

    # Phonopy harmonic calculation related methods
    def get_eigenvectors(self):
        if self._eigenvectors is None:
            # print("Getting frequencies & eigenvectors from Phonopy")
            self._eigenvectors, self._frequencies = (
                pho_interface.obtain_eigenvectors_and_frequencies(self.dynamic.structure,
                                                                  self.parameters.reduced_q_vector))
        return self._eigenvectors

    def get_frequencies(self):
        if self._frequencies is None:
            # print("Getting frequencies & eigenvectors from Phonopy")
            self._eigenvectors, self._frequencies = (
                pho_interface.obtain_eigenvectors_and_frequencies(self.dynamic.structure,
                                                                  self.parameters.reduced_q_vector))
        return self._frequencies

    def set_band_ranges(self, band_ranges):
        self.bands_clear()
        if isinstance(band_ranges, dict):
            self.parameters.band_ranges = band_ranges
        elif isinstance(band_ranges, list):
            self.parameters.band_ranges = {'ranges': band_ranges}
        else:
            raise Exception('Incorrect band ranges format')

    def get_band_ranges_and_labels(self):
        # return self.parameters.band_ranges
        if self.parameters.band_ranges is None:
            self.parameters.band_ranges = self.dynamic.structure.get_path_using_seek_path()

        return self.parameters.band_ranges

    def plot_phonon_dispersion_bands(self):
        bands_and_labels = self.get_band_ranges_and_labels()

        band_ranges = bands_and_labels['ranges']

        if self._bands is None:
            self._bands = pho_interface.obtain_phonon_dispersion_bands(self.dynamic.structure,
                                                                       band_ranges,
                                                                       NAC=self.parameters.use_NAC)

        for i, freq in enumerate(self._bands[1]):
            plt.plot(self._bands[1][i], self._bands[2][i], color='r')

            # plt.axes().get_xaxis().set_visible(False)
        # plt.axes().get_xaxis().set_ticks([])

        plt.ylabel('Frequency [THz]')
        plt.xlabel('Wave vector')
        plt.xlim([0, self._bands[1][-1][-1]])
        plt.axhline(y=0, color='k', ls='dashed')
        plt.suptitle('Phonon dispersion')

        if 'labels' in bands_and_labels:
            plt.rcParams.update({'mathtext.default': 'regular'})

            labels = bands_and_labels['labels']

            labels_e = []
            x_labels = []
            for i, freq in enumerate(self._bands[1]):
                if labels[i][0] == labels[i - 1][1]:
                    labels_e.append(replace_list(labels[i][0]))
                else:
                    labels_e.append(
                        replace_list(labels[i - 1][1]) + '/' + replace_list(labels[i][0]))
                x_labels.append(self._bands[1][i][0])
            x_labels.append(self._bands[1][-1][-1])
            labels_e.append(replace_list(labels[-1][1]))
            labels_e[0] = replace_list(labels[0][0])

            plt.xticks(x_labels, labels_e, rotation='horizontal')

        plt.show()

    def plot_renormalized_phonon_dispersion_bands(self, plot_linewidths=False, plot_harmonic=True):

        bands_full_data = self.get_renormalized_phonon_dispersion_bands(with_linewidths=plot_linewidths)

        plot_title = 'Renormalized phonon dispersion relations'
        for i, path in enumerate(bands_full_data):

            plt.plot(path['q_path_distances'], np.array(list(path['renormalized_frequencies'].values())).T, color='r',
                     label='Renormalized')

            if plot_harmonic:
                plt.plot(path['q_path_distances'], np.array(list(path['harmonic_frequencies'].values())).T, color='b',
                         label='Harmonic')

            if plot_linewidths:
                for freq, linewidth in zip(list(path['renormalized_frequencies'].values()),
                                           list(path['linewidth'].values())):
                    plt.fill_between(path['q_path_distances'], freq + np.array(linewidth) / 2,
                                     freq - np.array(linewidth) / 2, color='r', alpha=0.2, interpolate=True,
                                     linewidth=0)
                    plot_title = 'Renormalized phonon dispersion relations and linewidths'

        # plt.axes().get_xaxis().set_visible(False)
        plt.suptitle(plot_title)
        # plt.axes().get_xaxis().set_ticks([])
        plt.ylabel('Frequency [THz]')
        plt.xlabel('Wave vector')
        plt.xlim([0, bands_full_data[-1]['q_path_distances'][-1]])
        plt.axhline(y=0, color='k', ls='dashed')

        if plot_harmonic:
            try:  # Handle issues with old versions of matplotlib
                handles = plt.gca().get_legend_handles_labels()[0]
                plt.legend([handles[-1], handles[0]], ['Harmonic', 'Renormalized'])
            except IndexError:
                pass

        if 'labels' in bands_full_data[0]:
            plt.rcParams.update({'mathtext.default': 'regular'})

            labels = [[bands_full_data[i]['labels']['inf'],
                       bands_full_data[i]['labels']['sup']]
                      for i in range(len(bands_full_data))]

            labels_e = []
            x_labels = []
            for i, freq in enumerate(bands_full_data):
                if labels[i][0] == labels[i - 1][1]:
                    labels_e.append(replace_list(labels[i][0]))
                else:
                    labels_e.append(
                        replace_list(labels[i - 1][1]) + '/' + replace_list(labels[i][0]))
                x_labels.append(bands_full_data[i]['q_path_distances'][0])
            x_labels.append(bands_full_data[-1]['q_path_distances'][-1])
            labels_e.append(replace_list(labels[-1][1]))
            labels_e[0] = replace_list(labels[0][0])

            plt.xticks(x_labels, labels_e, rotation='horizontal')

        plt.show()

    def plot_linewidths_and_shifts_bands(self):

        bands_full_data = self.get_renormalized_phonon_dispersion_bands(with_linewidths=True,
                                                                        band_connection=True,
                                                                        interconnect_bands=True)

        number_of_branches = len(bands_full_data[0]['linewidth'])
        # print('number_of branches', number_of_branches)

        for i, path in enumerate(bands_full_data):
            prop_cicle = plt.rcParams['axes.prop_cycle']
            colors = prop_cicle.by_key()['color']

            for j in range(number_of_branches):
                plt.figure(0)
                branch = path['linewidth']['branch_{}'.format(j)]
                plt.plot(path['q_path_distances'], branch, color=np.roll(colors, -j)[0], label='linewidth')

                plt.figure(1)
                branch = path['harmonic_frequencies']['branch_{}'.format(j)]
                plt.plot(path['q_path_distances'], branch, color=np.roll(colors, -j)[0], label='linewidth')

                plt.figure(2)
                branch = path['frequency_shifts']['branch_{}'.format(j)]
                plt.plot(path['q_path_distances'], branch, color=np.roll(colors, -j)[0], label='linewidth')

                plt.figure(3)
                branch = path['renormalized_frequencies']['branch_{}'.format(j)]
                plt.plot(path['q_path_distances'], branch, color=np.roll(colors, -j)[0], label='linewidth')


        plt.figure(0)
        plt.suptitle('Phonon linewidths')

        plt.figure(1)
        plt.suptitle('Harmonic phonon dispersion relations')

        plt.figure(2)
        plt.suptitle('Frequency shifts')

        plt.figure(3)
        plt.suptitle('Renormalized phonon dispersion relations')

        for ifig in [0, 1, 2, 3]:
            plt.figure(ifig)

            plt.ylabel('Frequency [THz]')
            plt.xlabel('Wave vector')
            plt.xlim([0, bands_full_data[-1]['q_path_distances'][-1]])
            plt.axhline(y=0, color='k', ls='dashed')

            if 'labels' in bands_full_data[0]:
                plt.rcParams.update({'mathtext.default': 'regular'})

                labels = [[bands_full_data[i]['labels']['inf'],
                           bands_full_data[i]['labels']['sup']]
                          for i in range(len(bands_full_data))]

                labels_e = []
                x_labels = []
                for i, freq in enumerate(bands_full_data):
                    if labels[i][0] == labels[i - 1][1]:
                        labels_e.append(replace_list(labels[i][0]))
                    else:
                        labels_e.append(
                        replace_list(labels[i - 1][1]) + '/' + replace_list(labels[i][0]))
                    x_labels.append(bands_full_data[i]['q_path_distances'][0])
                x_labels.append(bands_full_data[-1]['q_path_distances'][-1])
                labels_e.append(replace_list(labels[-1][1]))
                labels_e[0] = replace_list(labels[0][0])
                plt.xticks(x_labels, labels_e, rotation='horizontal')

        plt.show()

    def plot_frequencies_vs_linewidths(self):

        qpoints, multiplicity, frequencies, linewidths = self.get_mesh_frequencies_and_linewidths()

        plt.ylabel('Linewidth [THz]')
        plt.xlabel('Frequency [THz]')

        plt.axhline(y=0, color='k', ls='dashed')
        plt.title('Frequency vs linewidths (from mesh: {})'.format(self.parameters.mesh_phonopy))
        for f, l in zip(np.array(frequencies).T, np.array(linewidths).T):
            plt.scatter(f, l)
        plt.show()

    def get_renormalized_phonon_dispersion_bands(self,
                                                 with_linewidths=False,
                                                 interconnect_bands=False,
                                                 band_connection=False):

        def reconnect_eigenvectors(bands):
            order = range(bands[2][0].shape[1])
            for i, ev_bands in enumerate(bands[3]):
                if i > 0:
                    ref = bands[3][i-1][-1]
                    metric = np.abs(np.dot(ref.conjugate().T, ev_bands[0]))
                    order = np.argmax(metric, axis=1)
                    bands[2][i] = bands[2][i].T[order].T
                    bands[3][i] = bands[3][i].T[order].T

        def reconnect_frequencies(bands):
            order = range(bands[2][0].shape[1])
            for i, f_bands in enumerate(bands[2]):
                if i > 0:
                    order = []
                    ref = np.array(bands[2][i-1][-1]).copy()
                    for j, test_val in enumerate(f_bands[0]):
                        ov = np.argmin(np.abs(ref - test_val))
                        order.append(ov)
                        ref[ov] = 1000

                    # print(order)
                    bands[2][i] = bands[2][i].T[order].T
                    bands[3][i] = bands[3][i].T[order].T

        def eigenvector_order(ev_bands, ev_renormalized):
            metric = np.zeros_like(np.abs(np.dot(ev_bands[0].conjugate().T, ev_renormalized[0])))
            for ev_ref, ev in zip(ev_bands, ev_renormalized):
                metric += np.abs(np.dot(ev_ref.conjugate().T, ev))
            order = np.argmax(metric, axis=1,)
            return order

        def set_order(bands, renormalized_bands):
            order_list = []
            for ev_bands, ev_renormalized in zip(bands[3], renormalized_bands[3]):
                order = eigenvector_order(ev_bands, ev_renormalized)
                order_list.append(order)

            freq = np.array(renormalized_bands[2]).copy()
            for i, o in enumerate(order_list):
                renormalized_bands[2][i] = freq[i].T[o].T

        renormalized_force_constants = self.get_renormalized_force_constants()
        bands = self.get_band_ranges_and_labels()
        band_ranges = bands['ranges']

        _bands = pho_interface.obtain_phonon_dispersion_bands(self.dynamic.structure,
                                                              band_ranges,
                                                              NAC=self.parameters.use_NAC,
                                                              band_connection=band_connection,
                                                              band_resolution=self.parameters.band_resolution)

        if interconnect_bands:
            # reconnect_frequencies(_bands)
            reconnect_eigenvectors(_bands)

        _renormalized_bands = pho_interface.obtain_phonon_dispersion_bands(self.dynamic.structure,
                                                                           band_ranges,
                                                                           force_constants=renormalized_force_constants,
                                                                           NAC=self.parameters.use_NAC,
                                                                           band_connection=band_connection,
                                                                           band_resolution=self.parameters.band_resolution)
        if band_connection:
            set_order(_bands, _renormalized_bands)

        data = self.get_commensurate_points_data()
        renormalized_frequencies = data['frequencies']
        eigenvectors = data['eigenvectors']
        linewidths = data['linewidths']
        fc_supercell = data['fc_supercell']

        sup_lim = pho_interface.get_renormalized_force_constants(renormalized_frequencies + linewidths / 2,
                                                                 eigenvectors,
                                                                 self.dynamic.structure,
                                                                 fc_supercell,
                                                                 symmetrize=self.parameters.symmetrize)

        inf_lim = pho_interface.get_renormalized_force_constants(renormalized_frequencies - linewidths / 2,
                                                                 eigenvectors,
                                                                 self.dynamic.structure,
                                                                 fc_supercell,
                                                                 symmetrize=self.parameters.symmetrize)

        if with_linewidths:
            renormalized_bands_s = pho_interface.obtain_phonon_dispersion_bands(self.dynamic.structure,
                                                                                band_ranges,
                                                                                force_constants=sup_lim,
                                                                                NAC=self.parameters.use_NAC,
                                                                                band_connection=band_connection,
                                                                                band_resolution=self.parameters.band_resolution)

            renormalized_bands_i = pho_interface.obtain_phonon_dispersion_bands(self.dynamic.structure,
                                                                                band_ranges,
                                                                                force_constants=inf_lim,
                                                                                NAC=self.parameters.use_NAC,
                                                                                band_connection=band_connection,
                                                                                band_resolution=self.parameters.band_resolution)

            if band_connection:
                set_order(_bands, renormalized_bands_s)
                set_order(_bands, renormalized_bands_i)

        bands_full_data = []
        for i, q_path in enumerate(_bands[1]):

            band = {'q_path_distances': q_path.tolist(),
                    'q_bounds': {'inf': list(band_ranges[i][0]), 'sup': list(band_ranges[i][1])},
                    'harmonic_frequencies': {'branch_{}'.format(key): value.tolist() for (key, value) in
                                             enumerate(_bands[2][i].T)},
                    'renormalized_frequencies': {'branch_{}'.format(key): value.tolist() for (key, value) in
                                                 enumerate(_renormalized_bands[2][i].T)},
                    'frequency_shifts': {'branch_{}'.format(key): value.tolist() for (key, value) in
                                         enumerate(_renormalized_bands[2][i].T - _bands[2][i].T)},
                    }

            if with_linewidths:
                band.update({'linewidth_minus': {'branch_{}'.format(key): value.tolist() for (key, value) in
                                                 enumerate(renormalized_bands_i[2][i].T)},
                             'linewidth_plus': {'branch_{}'.format(key): value.tolist() for (key, value) in
                                                enumerate(renormalized_bands_s[2][i].T)},
                             'linewidth': {'branch_{}'.format(key): value.tolist() for (key, value) in
                                           enumerate(renormalized_bands_s[2][i].T - renormalized_bands_i[2][i].T)}}
                            )

            if 'labels' in bands:
                labels = bands['labels']
                band.update({'labels': {'inf': labels[i][0], 'sup': labels[i][1]}})

            bands_full_data.append(band)

        return bands_full_data

    def get_mesh_frequencies_and_linewidths(self):

        data = self.get_commensurate_points_data()

        renormalized_frequencies = data['frequencies']
        eigenvectors = data['eigenvectors']
        linewidths = data['linewidths']
        fc_supercell = data['fc_supercell']

        linewidths_fc = pho_interface.get_renormalized_force_constants(linewidths,
                                                                   eigenvectors,
                                                                   self.dynamic.structure,
                                                                   fc_supercell,
                                                                   symmetrize=self.parameters.symmetrize)

        _, _, linewidths_mesh = pho_interface.obtain_phonopy_mesh_from_force_constants(self.dynamic.structure,
                                                                                       force_constants=linewidths_fc,
                                                                                       mesh=self.parameters.mesh_phonopy,
                                                                                       NAC=None)

        frequencies_fc = pho_interface.get_renormalized_force_constants(renormalized_frequencies,
                                                                        eigenvectors,
                                                                        self.dynamic.structure,
                                                                        fc_supercell,
                                                                        symmetrize=self.parameters.symmetrize)

        qpoints, multiplicity, frequencies_mesh = pho_interface.obtain_phonopy_mesh_from_force_constants(self.dynamic.structure,
                                                                                        force_constants=frequencies_fc,
                                                                                        mesh=self.parameters.mesh_phonopy,
                                                                                        NAC=None)

        return qpoints, multiplicity, frequencies_mesh, linewidths_mesh

    def write_renormalized_phonon_dispersion_bands(self, filename='bands_data.yaml'):
        bands_full_data = self.get_renormalized_phonon_dispersion_bands(with_linewidths=True, band_connection=True)
        reading.save_bands_data_to_file(bands_full_data, filename)

    def print_phonon_dispersion_bands(self):
        if self._bands is None:
            self._bands = pho_interface.obtain_phonon_dispersion_bands(self.dynamic.structure,
                                                                       self.get_band_ranges_and_labels(),
                                                                       NAC=self.parameters.use_NAC)
        np.set_printoptions(linewidth=200)
        for i, freq in enumerate(self._bands[1]):
            print(str(np.hstack([self._bands[1][i][None].T, self._bands[2][i]])).replace('[', '').replace(']', ''))

    def plot_eigenvectors(self):
        modes.plot_phonon_modes(self.dynamic.structure,
                                self.get_eigenvectors(),
                                self.get_q_vector(),
                                vectors_scale=self.parameters.modes_vectors_scale)

    def plot_dos_phonopy(self, force_constants=None):

        phonopy_dos = pho_interface.obtain_phonopy_dos(self.dynamic.structure,
                                                       mesh=self.parameters.mesh_phonopy,
                                                       projected_on_atom=self.parameters.project_on_atom,
                                                       NAC=self.parameters.use_NAC)

        plt.plot(phonopy_dos[0], phonopy_dos[1], 'b-', label='Harmonic')

        if force_constants is not None:
            phonopy_dos_r = pho_interface.obtain_phonopy_dos(self.dynamic.structure,
                                                             mesh=self.parameters.mesh_phonopy,
                                                             force_constants=force_constants,
                                                             projected_on_atom=self.parameters.project_on_atom,
                                                             NAC=self.parameters.use_NAC)

            plt.plot(phonopy_dos_r[0], phonopy_dos_r[1], 'g-', label='Renormalized')

        plt.title('Density of states (Normalized to unit cell)')
        plt.xlabel('Frequency [THz]')
        plt.ylabel('Density of states')
        plt.legend()
        plt.axhline(y=0, color='k', ls='dashed')
        plt.show()

    def check_commensurate(self, q_point, decimals=4):
        supercell = self.dynamic.get_supercell_matrix()

        commensurate = False
        primitive_matrix = self.dynamic.structure.get_primitive_matrix()

        transform = np.dot(q_point, np.linalg.inv(primitive_matrix))
        transform = np.multiply(transform, supercell)
        transform = np.around(transform, decimals=decimals)

        if np.all(np.equal(np.mod(transform, 1), 0)):
            commensurate = True

        return commensurate

    # Projections related methods
    def get_vc(self):
        if self._vc is None:
            print("Projecting into wave vector")
            # Check if commensurate point
            if not self.check_commensurate(self.get_reduced_q_vector()):
                print("warning! This wave vector is not a commensurate q-point in MD supercell")

            if self.parameters.project_on_atom > -1:
                element = self.dynamic.structure.get_atomic_elements(unique=True)[self.parameters.project_on_atom]
                print('Project on atom {} : {}'.format(self.parameters.project_on_atom, element))

            self._vc = projection.project_onto_wave_vector(self.dynamic,
                                                           self.get_q_vector(),
                                                           project_on_atom=self.parameters.project_on_atom)
        return self._vc

    def get_vq(self):
        if self._vq is None:
            print("Projecting into phonon mode")
            self._vq = projection.project_onto_phonon(self.get_vc(), self.get_eigenvectors())
        return self._vq

    def plot_vq(self, modes=None):
        if not modes:
            modes = [0]
        plt.suptitle('Phonon mode projection')
        plt.xlabel('Time [ps]')
        plt.ylabel(r'$u^{1/2}\AA/ps$')

        time = np.linspace(0, self.get_vc().shape[0] * self.dynamic.get_time_step_average(),
                           num=self.get_vc().shape[0])

        for mode in modes:
            plt.plot(time, self.get_vq()[:, mode].real, label='mode: ' + str(mode))
        plt.legend()
        plt.show()

    def plot_vc(self, atoms=None, coordinates=None):
        if not atoms:
            atoms = [0]
        if not coordinates:
            coordinates = [0]
        time = np.linspace(0, self.get_vc().shape[0] * self.dynamic.get_time_step_average(),
                           num=self.get_vc().shape[0])

        plt.suptitle('Wave vector projection')
        plt.xlabel('Time [ps]')
        plt.ylabel(r'$u^{1/2}\AA/ps$')

        for atom in atoms:
            for coordinate in coordinates:
                plt.plot(time,
                         self.get_vc()[:, atom, coordinate].real,
                         label='atom: ' + str(atom) + ' coordinate:' + str(coordinate))
        plt.legend()
        plt.show()

    def save_vc(self, file_name, atom=1):
        print("Saving wave vector projection to file")
        if not 0 < atom <= self.get_vc().shape[1]:
            raise Exception('Atom number {} does not exist in primitive cell'.format(atom))
        # np.savetxt(file_name, self.get_vc()[:, atom, :].real)
        np.savetxt(file_name, self.get_vc()[:, atom-1, :], fmt='%+.8e%+.8ej    '*3)

    def save_vq(self, file_name):
        ndim = self.get_vq().shape[1]
        print("Saving phonon projection to file")
        # np.savetxt(file_name, self.get_vq())
        np.savetxt(file_name, self.get_vq(), fmt='%+.8e%+.8ej    '*ndim)

    # Power spectra related methods
    def select_power_spectra_algorithm(self, algorithm):
        if algorithm in power_spectrum_functions.keys():
            if algorithm != self.parameters.power_spectra_algorithm:
                self.power_spectra_clear()
                self.parameters.power_spectra_algorithm = algorithm
            print("Using {0} function".format(power_spectrum_functions[algorithm][1]))
        else:
            print("Power spectrum algorithm number not found!\nPlease select:")
            for i in power_spectrum_functions.keys():
                print('{0} : {1}'.format(i, power_spectrum_functions[i][1]))
            exit()

    def select_fitting_function(self, function):
        from macer.externals.dynaphopy_bundled.analysis.fitting.fitting_functions import fitting_functions
        if function in fitting_functions.keys():
            if function != self.parameters.fitting_function:
                self.force_constants_clear()
                self.parameters.fitting_function = function
        else:
            print("Fitting function number not found!\nPlease select:")
            for i in fitting_functions.keys():
                print('{0} : {1}'.format(i, fitting_functions[i]))
            exit()

    def get_power_spectrum_phonon(self):
        if self._power_spectrum_phonon is None:
            print("Calculating phonon projection power spectra")

            if self.parameters.use_symmetry:
                initial_reduced_q_point = self.get_reduced_q_vector()
                power_spectrum_phonon = []
                q_points_equivalent = pho_interface.get_equivalent_q_points_by_symmetry(self.get_reduced_q_vector(),
                                                                                        self.dynamic.structure)
                #                print(q_points_equivalent)
                for q_point in q_points_equivalent:
                    self.set_reduced_q_vector(q_point)
                    power_spectrum_phonon.append(
                        (power_spectrum_functions[self.parameters.power_spectra_algorithm])[0](self.get_vq(),
                                                                                               self.dynamic,
                                                                                               self.parameters))

                self.set_reduced_q_vector(initial_reduced_q_point)
                self._power_spectrum_phonon = np.average(power_spectrum_phonon, axis=0)
            else:
                self._power_spectrum_phonon = (
                    power_spectrum_functions[self.parameters.power_spectra_algorithm])[0](self.get_vq(),
                                                                                          self.dynamic,
                                                                                          self.parameters)

        return self._power_spectrum_phonon

    def get_power_spectrum_wave_vector(self):

        if self._power_spectrum_wave_vector is None:
            print('Calculating wave vector projection power spectrum')
            size = self.get_vc().shape[1] * self.get_vc().shape[2]
            if self.parameters.use_symmetry:
                initial_reduced_q_point = self.get_reduced_q_vector()
                power_spectrum_wave_vector = []
                q_points_equivalent = pho_interface.get_equivalent_q_points_by_symmetry(self.get_reduced_q_vector(),
                                                                                        self.dynamic.structure)
                #  print(q_points_equivalent)
                for q_point in q_points_equivalent:
                    self.set_reduced_q_vector(q_point)

                    power_spectrum_wave_vector.append((power_spectrum_functions[
                                                           self.parameters.power_spectra_algorithm])[0](
                        self.get_vc().swapaxes(1, 2).reshape(-1, size),
                        self.dynamic,
                        self.parameters))
                power_spectrum_wave_vector = np.array(power_spectrum_wave_vector)
                self.set_reduced_q_vector(initial_reduced_q_point)

                self._power_spectrum_wave_vector = np.average(power_spectrum_wave_vector, axis=0)

            else:
                self._power_spectrum_wave_vector = (
                    power_spectrum_functions[self.parameters.power_spectra_algorithm])[0](
                    self.get_vc().swapaxes(1, 2).reshape(-1, size),
                    self.dynamic,
                    self.parameters)

        return np.nansum(self._power_spectrum_wave_vector, axis=1)

    def get_power_spectrum_full(self, projection_on_coordinate=-1):

        # temporal interface
        number_of_dimensions = self.dynamic.structure.get_number_of_dimensions()
        projected_atom_type = self.parameters.project_on_atom

        if self._power_spectrum_direct is None:
            print("Calculation full power spectrum")

            velocity_mass_average = self.dynamic.get_velocity_mass_average()

            if projected_atom_type >= 0:
                print('Power spectrum projected onto atom type {0}'.format(projected_atom_type))
                supercell = self.dynamic.get_supercell_matrix()
                atom_types = np.array(self.dynamic.structure.get_atom_type_index(supercell=supercell))
                atom_indices = np.argwhere(atom_types == projected_atom_type).flatten()
                if len(atom_indices) == 0:
                    print('Atom type {0} does not exist'.format(projected_atom_type))
                    exit()

                # Only works if project on atom is requested!
                if projection_on_coordinate >= number_of_dimensions:
                    print('Projected coordinate should be smaller than {}'.format(number_of_dimensions))
                    exit()
                if projection_on_coordinate > -1:
                    print('Power spectrum projected onto coordinate {}'.format(projection_on_coordinate))
                    velocity_mass_average = velocity_mass_average[:, atom_indices, projection_on_coordinate, None]
                else:
                    velocity_mass_average = velocity_mass_average[:, atom_indices]

            size = velocity_mass_average.shape[1] * velocity_mass_average.shape[2]

            # Memory efficient algorithm
            if self.parameters.silent:
                self._power_spectrum_direct = np.zeros_like(self.parameters.frequency_range[None].T)
                for i in range(velocity_mass_average.shape[1]):
                    for j in range(velocity_mass_average.shape[2]):
                        self._power_spectrum_direct += \
                            (power_spectrum_functions[self.parameters.power_spectra_algorithm])[0](
                                velocity_mass_average[:, i, j][None].T,
                                self.dynamic,
                                self.parameters)

            else:
                self._power_spectrum_direct = (power_spectrum_functions[self.parameters.power_spectra_algorithm])[0](
                    velocity_mass_average.swapaxes(1, 2).reshape(-1, size),
                    self.dynamic,
                    self.parameters)

            self._power_spectrum_direct = np.sum(self._power_spectrum_direct, axis=1)
        return self._power_spectrum_direct

    def get_power_spectrum_partials(self, save_to_file=None):

        if self._power_spectrum_partials is None:
            print("Calculation power spectrum partials")

            velocity_mass_average = self.dynamic.get_velocity_mass_average()

            self._power_spectrum_partials = (power_spectrum_functions[self.parameters.power_spectra_algorithm])[0](
                velocity_mass_average[:, :, 0],
                self.dynamic,
                self.parameters)

            for i in [1, 2]:
                self._power_spectrum_partials += (power_spectrum_functions[self.parameters.power_spectra_algorithm])[0](
                    velocity_mass_average[:, :, i],
                    self.dynamic,
                    self.parameters)

        if save_to_file is not None:
            np.savetxt(save_to_file, np.hstack([self.get_frequency_range()[None].T, self._power_spectrum_partials]))

        return self._power_spectrum_partials

    def phonon_width_scan_analysis(self):
        from macer.externals.dynaphopy_bundled.power_spectrum import mem_coefficient_scan_analysis
        print("Phonon coefficient scan analysis(Maximum entropy method/Symmetric Lorentzian fit only)")
        mem_coefficient_scan_analysis(self.get_vq(), self.dynamic, self.parameters)

    def phonon_individual_analysis(self):
        print("Peak analysis")

        fitting.phonon_fitting_analysis(self.get_power_spectrum_phonon(),
                                        self.parameters.frequency_range,
                                        harmonic_frequencies=self.get_frequencies(),
                                        thermal_expansion_shift=self.get_qha_shift(self.get_reduced_q_vector()),
                                        show_plots=not self.parameters.silent,
                                        fitting_function_type=self.parameters.fitting_function,
                                        use_degeneracy=self.parameters.use_symmetry,
                                        show_occupancy=self.parameters.project_on_atom < 0  # temporal interface
                                        )
        return

    def plot_power_spectrum_full(self):

        fig, ax1 = plt.subplots()

        ax1.plot(self.get_frequency_range(), self.get_power_spectrum_full(), 'r-', label='Power spectrum (MD)')
        ax1.set_xlabel('Frequency [THz]')
        ax1.set_ylabel('eV * ps')
        ax2 = ax1.twinx()

        if self.dynamic.structure.forces_available():
            phonopy_dos = pho_interface.obtain_phonopy_dos(self.dynamic.structure,
                                                           mesh=self.parameters.mesh_phonopy,
                                                           freq_min=self.get_frequency_range()[0],
                                                           freq_max=self.get_frequency_range()[-1],
                                                           freq_pitch=self.parameters.spectrum_resolution,
                                                           projected_on_atom=self.parameters.project_on_atom,
                                                           NAC=self.parameters.use_NAC)

            ax2.plot(phonopy_dos[0], phonopy_dos[1], 'b-', label='DoS (Lattice dynamics)')
            ax2.set_ylabel('Density of states')

        if self._renormalized_force_constants is not None:
            phonopy_dos_r = pho_interface.obtain_phonopy_dos(self.dynamic.structure,
                                                             mesh=self.parameters.mesh_phonopy,
                                                             freq_min=self.get_frequency_range()[0],
                                                             freq_max=self.get_frequency_range()[-1],
                                                             freq_pitch=self.parameters.spectrum_resolution,
                                                             force_constants=self._renormalized_force_constants,
                                                             projected_on_atom=self.parameters.project_on_atom,
                                                             NAC=self.parameters.use_NAC)

            ax2.plot(phonopy_dos_r[0], phonopy_dos_r[1], 'g-', label='Renormalized DoS')

        plt.suptitle('Full power spectrum')

        handles1, labels = ax1.get_legend_handles_labels()
        handles2, labels = ax2.get_legend_handles_labels()

        handles = handles1 + handles2
        plt.legend(handles, ['Power spectrum (MD)', 'DoS (Harmonic)', 'DoS (Renormalized)'])
        # plt.legend()
        plt.show()

        total_integral = simps(self.get_power_spectrum_full(), x=self.get_frequency_range())
        print("Total Area (Kinetic energy <K>): {0} eV".format(total_integral))

    def plot_power_spectrum_wave_vector(self):
        plt.suptitle('Projection onto wave vector')
        plt.plot(self.get_frequency_range(), self.get_power_spectrum_wave_vector(), 'r-')
        plt.xlabel('Frequency [THz]')
        plt.ylabel('eV * ps')
        plt.axhline(y=0, color='k', ls='dashed')
        plt.show()
        total_integral = simps(self.get_power_spectrum_wave_vector(), x=self.get_frequency_range())
        print("Total Area (Kinetic energy <K>): {0} eV".format(total_integral))

    def plot_power_spectrum_phonon(self):
        for i in range(self.get_power_spectrum_phonon().shape[1]):
            plt.figure(i)
            plt.suptitle('Projection onto phonon mode {0}'.format(i + 1))
            plt.plot(self.get_frequency_range(), self.get_power_spectrum_phonon()[:, i])
            plt.xlabel('Frequency [THz]')
            plt.ylabel('eV * ps')
            plt.axhline(y=0, color='k', ls='dashed')

        plt.show()

    # Plot dynamical properties related methods
    def plot_trajectory(self, atoms=None, coordinates=None):
        if atoms is None: atoms = [0]
        if coordinates is None: coordinates = [0]

        plt.suptitle('Trajectory')
        time = np.linspace(0, self.dynamic.trajectory.shape[0] * self.dynamic.get_time_step_average(),
                           num=self.dynamic.trajectory.shape[0])
        for atom in atoms:
            for coordinate in coordinates:
                plt.plot(time, self.dynamic.trajectory[:, atom, coordinate].real,
                         label='atom: {0}  coordinate: {1}'.format(atom, coordinate))

        plt.legend()
        plt.xlabel('Time [ps]')
        plt.ylabel('Angstrom')
        plt.show()

    def plot_velocity(self, atoms=None, coordinates=None):
        if not atoms: atoms = [0]
        if not coordinates: coordinates = [0]

        plt.suptitle('Velocity')
        time = np.linspace(0, self.dynamic.velocity.shape[0] * self.dynamic.get_time_step_average(),
                           num=self.dynamic.velocity.shape[0])

        for atom in atoms:
            for coordinate in coordinates:
                plt.plot(time, self.dynamic.velocity[:, atom, coordinate].real,
                         label='atom: {0}  coordinate: {1}'.format(atom, coordinate))

        plt.legend()
        plt.xlabel('Time [ps]')
        plt.ylabel(r'$\AA/ps$')
        plt.show()

    def plot_energy(self):
        plt.suptitle('Energy')
        plt.plot(self.dynamic.get_time().real,
                 self.dynamic.get_energy().real)
        plt.show()

    def plot_trajectory_distribution(self, direction):
        from macer.externals.dynaphopy_bundled.analysis.fitting.fitting_functions import Gaussian_function

        atomic_types = self.dynamic.structure.get_atomic_elements()
        atom_type_index_unique = np.unique(self.dynamic.structure.get_atom_type_index(), return_index=True)[1]
        atomic_types_unique = [atomic_types[i] for i in atom_type_index_unique]

        direction = np.array(direction)

        distributions, distance = self.get_atomic_displacements(direction)

        plt.figure()
        for atom in range(distributions.shape[0]):

            plt.figure(atom + 1)
            plt.title('Atomic displacements')
            plt.suptitle('Atom {0}, Element {1}'.format(atom, atomic_types_unique[atom]))

            width = (distance[1] - distance[0])
            center = (distance[:-1] + distance[1:] + width) / 2

            print('\nAtom {0}, Element {1}'.format(atom, atomic_types_unique[atom]))
            print('-----------------------------------------')
            try:

                distance_centers = distance[:-1] + width
                fitting_function = Gaussian_function(distance_centers,
                                                     distributions[atom],
                                                     guess_height=1,
                                                     guess_position=0)

                parameters = fitting_function.get_fitting()
                print('Mean               {0:15.6f} Angstrom'.format(parameters['peak_position']))
                print('Standard deviation {0:15.6f} Angstrom'.format(parameters['width']))
                print('Global fit error   {0:15.6f}'.format(parameters['global_error']))

                plt.plot(distance, fitting_function.get_curve(distance),
                         label=fitting_function.curve_name,
                         linewidth=3, color='g')
            except:
                print('Gaussian fitting failed')

            plt.bar(center, distributions[atom], align='center', width=width)
            plt.xlabel('Direction: ' + ' '.join(np.array(direction, dtype=str)) + ' [Angstrom]')
            plt.xlim([distance[0], distance[-1]])
            plt.ylim([0, None])
            plt.axhline(y=0, color='k', ls='dashed')
            plt.legend()
        plt.show()

    # Printing data to files
    def write_power_spectrum_full(self, file_name):
        reading.write_curve_to_file(self.get_frequency_range(),
                                    self.get_power_spectrum_full()[None].T,
                                    file_name)
        total_integral = simps(self.get_power_spectrum_full(), x=self.get_frequency_range())
        print("Total Area (Kinetic energy <K>): {0} eV".format(total_integral))

    def write_power_spectrum_wave_vector(self, file_name):
        reading.write_curve_to_file(self.get_frequency_range(),
                                    self.get_power_spectrum_wave_vector()[None].T,
                                    file_name)
        total_integral = simps(self.get_power_spectrum_wave_vector(), x=self.get_frequency_range())
        print("Total Area (Kinetic energy <K>): {0} eV".format(total_integral))

    def write_power_spectrum_phonon(self, file_name):
        reading.write_curve_to_file(self.get_frequency_range(),
                                    self.get_power_spectrum_phonon(),
                                    file_name)

    def get_atomic_displacements(self, direction):

        number_of_bins = self.parameters.number_of_bins_histogram
        direction = np.array(direction)

        projections = trajdist.trajectory_projection(self.dynamic, direction)

        min_val = np.amin(projections)
        max_val = np.amax(projections)

        bins = None
        distributions = []
        for atom in range(projections.shape[0]):
            distribution, bins = np.histogram(projections[atom],
                                              bins=number_of_bins,
                                              range=(min_val, max_val),
                                              density=True)

            distributions.append(distribution)

        distance = np.array([i_bin - (bins[1] - bins[0]) / 2 for i_bin in bins])

        return np.array(distributions), distance

    def write_atomic_displacements(self, direction, file_name):
        distributions, distance = self.get_atomic_displacements(direction)
        reading.write_curve_to_file(distance, distributions.T, file_name)

    # Molecular dynamics analysis related methods
    def show_boltzmann_distribution(self):
        energy.boltzmann_distribution(self.dynamic, self.parameters)

    def get_temperature(self):
        if not self._temperature:
            save_status = self.parameters.silent
            self.parameters.silent = True
            self._temperature = energy.boltzmann_distribution(self.dynamic, self.parameters)
            self.parameters.silent = save_status
        return self._temperature

    def set_temperature(self, temperature):
        self._temperature = temperature

    def get_algorithm_list(self):
        return power_spectrum_functions.values()

    def get_commensurate_points_data(self, auto_range=True):

        if self._commensurate_points_data is None:

            if auto_range:
                # Get range from harmonic DOS
                phonopy_dos = pho_interface.obtain_phonopy_dos(self.dynamic.structure,
                                                               mesh=self.parameters.mesh_phonopy)

                self.set_frequency_limits([0, np.max(phonopy_dos[0][-1]) * 1.2])
                print('set frequency range: {} - {}'.format(self.get_frequency_range()[0],
                                                            self.get_frequency_range()[-1]))

            # Decide the size of the supercell to use to calculate the renormalized force constants
            if self._parameters.use_MD_cell_commensurate:
                fc_supercell = np.diag(self.dynamic.get_supercell_matrix())
            else:
                fc_supercell = self.dynamic.structure.get_supercell_phonon()

            com_points = pho_interface.get_commensurate_points(self.dynamic.structure,
                                                               fc_supercell)

            initial_reduced_q_vector = self.get_reduced_q_vector()

            renormalized_frequencies = []
            frequency_shifts = []
            eigenvectors = []
            linewidths = []
            q_points_list = []

            for i, reduced_q_vector in enumerate(com_points):

                print("\nQ-point: {0} / {1} ".format(i + 1, len(com_points)) +
                      "    [{:8.5f} {:8.5f} {:8.5f} ]".format(*reduced_q_vector))

                self.set_reduced_q_vector(reduced_q_vector)
                eigenvectors.append(self.get_eigenvectors())

                q_points_equivalent = pho_interface.get_equivalent_q_points_by_symmetry(reduced_q_vector,
                                                                                        self.dynamic.structure)
                q_index = _vector_in_list(q_points_equivalent, q_points_list)
                q_points_list.append(reduced_q_vector)

                if q_index != 0 and self.parameters.use_symmetry:
                    renormalized_frequencies.append(renormalized_frequencies[q_index])
                    linewidths.append(linewidths[q_index])
                    frequency_shifts.append(frequency_shifts[q_index])

                    print('Skipped, equivalent to {0}'.format(q_points_list[q_index]))
                    continue

                self.set_reduced_q_vector(reduced_q_vector)

                data = fitting.phonon_fitting_analysis(self.get_power_spectrum_phonon(),
                                                       self.parameters.frequency_range,
                                                       harmonic_frequencies=self.get_frequencies(),
                                                       thermal_expansion_shift=self.get_qha_shift(reduced_q_vector),
                                                       show_plots=False,
                                                       fitting_function_type=self.parameters.fitting_function,
                                                       use_degeneracy=self.parameters.use_symmetry)

                positions = data['positions']
                widths = data['widths']
                if (reduced_q_vector == [0, 0, 0]).all():
                    print('Fixing gamma point 0 frequencies')
                    positions[0] = 0.
                    positions[1] = 0.
                    positions[2] = 0.
                    widths[0] = 0.
                    widths[1] = 0.
                    widths[2] = 0.

                renormalized_frequencies.append(positions)
                linewidths.append(widths)
                frequency_shifts.append(np.array(positions) - self.get_frequencies())

            renormalized_frequencies = np.array(renormalized_frequencies)
            linewidths = np.array(linewidths)
            frequency_shifts = np.array(frequency_shifts)

            # To be deprecated
            if self.parameters.save_renormalized_frequencies:
                print("This option will be deprecated in the future. Please use save quasiparticle data option")
                np.savetxt('renormalized_frequencies', renormalized_frequencies)
            # np.savetxt('test_line', linewidths)

            self._commensurate_points_data = {'frequencies': renormalized_frequencies,
                                              'eigenvectors': eigenvectors,
                                              'linewidths': linewidths,
                                              'frequency_shifts': frequency_shifts,
                                              'q_points': q_points_list,
                                              'fc_supercell': fc_supercell}

            self.set_reduced_q_vector(initial_reduced_q_vector)

        return self._commensurate_points_data

    def get_renormalized_force_constants(self):
        data = self.get_commensurate_points_data()
        renormalized_frequencies = data['frequencies']
        eigenvectors = data['eigenvectors']
        fc_supercell = data['fc_supercell']

        if self._renormalized_force_constants is None:
            self._renormalized_force_constants = pho_interface.get_renormalized_force_constants(
                renormalized_frequencies,
                eigenvectors,
                self.dynamic.structure,
                fc_supercell,
                symmetrize=self.parameters.symmetrize)

        return self._renormalized_force_constants

    def get_commensurate_points_properties(self):

        #  Decide the size of the supercell to use to calculate the renormalized force constants
        if self._parameters.use_MD_cell_commensurate:
            fc_supercell = np.diag(self.dynamic.get_supercell_matrix())
        else:
            fc_supercell = self.dynamic.structure.get_supercell_phonon()

        com_points = pho_interface.get_commensurate_points(self.dynamic.structure,
                                                           fc_supercell)

        group_velocity_list = []
        q_points_list = []
        for i, reduced_q_vector in enumerate(com_points):

            gv = pho_interface.obtain_phonopy_group_velocity(self.dynamic.structure,
                                                             reduced_q_vector,
                                                             force_constants=self.get_renormalized_force_constants())

            group_velocity_list.append(gv)
            q_points_list.append(reduced_q_vector)

        return {'group_velocity': group_velocity_list,
                'q_points': q_points_list,
                'fc_supercell': fc_supercell}

    def write_renormalized_constants(self, filename="FORCE_CONSTANTS"):
        force_constants = self.get_renormalized_force_constants()
        pho_interface.save_force_constants_to_file(force_constants, filename)

    def write_quasiparticles_data(self, filename="quasiparticles_data.yaml", with_extra=False):
        quasiparticle_data = self.get_commensurate_points_data()
        if with_extra:
            quasiparticle_data.update(self.get_commensurate_points_properties())
        reading.save_quasiparticle_data_to_file(quasiparticle_data, filename)

    def write_mesh_data(self, file_name='mesh_data.yaml'):
        mesh_data = self.get_mesh_frequencies_and_linewidths()
        reading.save_mesh_data_to_yaml_file(mesh_data, file_name)

    def get_thermal_properties(self, force_constants=None, normalize_dos=False):

        temperature = self.get_temperature()

        phonopy_dos = pho_interface.obtain_phonopy_dos(self.dynamic.structure,
                                                       mesh=self.parameters.mesh_phonopy,
                                                       freq_min=0.01,
                                                       freq_max=self.get_frequency_range()[-1],
                                                       freq_pitch=self.parameters.spectrum_resolution,
                                                       force_constants=force_constants,
                                                       projected_on_atom=self.parameters.project_on_atom,
                                                       NAC=self.parameters.use_NAC)

        integration = simps(phonopy_dos[1], x=phonopy_dos[0]) / (
                self.dynamic.structure.get_number_of_atoms() *
                self.dynamic.structure.get_number_of_dimensions())

        if normalize_dos:
            phonopy_dos[1] /= integration
            integration = 1.0
            if self.parameters.project_on_atom > -1:
                phonopy_dos[1] /= self.dynamic.structure.get_number_of_primitive_atoms()
                integration /= self.dynamic.structure.get_number_of_primitive_atoms()

        free_energy = thm.get_free_energy(temperature, phonopy_dos[0], phonopy_dos[1])
        entropy = thm.get_entropy(temperature, phonopy_dos[0], phonopy_dos[1])
        c_v = thm.get_cv(temperature, phonopy_dos[0], phonopy_dos[1])
        total_energy = thm.get_total_energy(temperature, phonopy_dos[0], phonopy_dos[1])

        if force_constants is not None:
            # Free energy correction
            phonopy_dos_h = pho_interface.obtain_phonopy_dos(self.dynamic.structure,
                                                             mesh=self.parameters.mesh_phonopy,
                                                             freq_min=0.01,
                                                             freq_max=self.get_frequency_range()[-1],
                                                             freq_pitch=self.parameters.spectrum_resolution,
                                                             projected_on_atom=self.parameters.project_on_atom,
                                                             NAC=self.parameters.use_NAC)

            free_energy += thm.get_free_energy_correction_dos(temperature, phonopy_dos_h[0], phonopy_dos_h[1],
                                                              phonopy_dos[1])
            total_energy += thm.get_free_energy_correction_dos(temperature, phonopy_dos_h[0], phonopy_dos_h[1],
                                                               phonopy_dos[1])

            # correction = thm.get_free_energy_correction_dos(temperature, phonopy_dos_h[0], phonopy_dos_h[1], phonopy_dos[1])
            # print('Free energy/total energy correction: {0:12.4f} KJ/mol'.format(correction))

        return [free_energy, entropy, c_v, total_energy, integration]

    def display_thermal_properties(self, from_power_spectrum=False, normalize_dos=False, print_phonopy=False):

        temperature = self.get_temperature()

        print('Using mesh: {0}'.format(self.parameters.mesh_phonopy))

        if print_phonopy:
            harmonic_properties = pho_interface.obtain_phonopy_thermal_properties(self.dynamic.structure,
                                                                                  temperature,
                                                                                  mesh=self.parameters.mesh_phonopy,
                                                                                  NAC=self.parameters.use_NAC)

            renormalized_properties = pho_interface.obtain_phonopy_thermal_properties(self.dynamic.structure,
                                                                                      temperature,
                                                                                      mesh=self.parameters.mesh_phonopy,
                                                                                      force_constants=self.get_renormalized_force_constants(),
                                                                                      NAC=self.parameters.use_NAC)

            print('\nThermal properties per unit cell ({0:.2f} K) [From phonopy (Reference)]\n'
                  '-----------------------------------------------------------------'.format(temperature))
            print('                                   Harmonic   Quasiparticle\n')
            print('Free energy (not corrected):   {0:12.4f}  {3:12.4f}   KJ/mol\n'
                  'Entropy:                       {1:12.4f}  {4:12.4f}   J/K/mol\n'
                  'Cv:                            {2:12.4f}  {5:12.4f}   J/K/mol\n'.format(
                *(harmonic_properties + renormalized_properties)))

        harmonic_properties = self.get_thermal_properties(normalize_dos=normalize_dos)
        renormalized_properties = self.get_thermal_properties(force_constants=self.get_renormalized_force_constants(),
                                                              normalize_dos=normalize_dos)
        frequency_range = self.get_frequency_range()
        resolution = frequency_range[1] - frequency_range[0]

        if from_power_spectrum:
            normalization = np.prod(self.dynamic.get_supercell_matrix())

            power_spectrum_dos = thm.get_dos(temperature, frequency_range, self.get_power_spectrum_full(),
                                             normalization)
            integration = simps(power_spectrum_dos, x=frequency_range) / (
                self.dynamic.structure.get_number_of_atoms() *
                self.dynamic.structure.get_number_of_dimensions())

            if normalize_dos:
                power_spectrum_dos /= integration
                integration = 1.0
                if self.parameters.project_on_atom > -1:
                    power_spectrum_dos /= self.dynamic.structure.get_number_of_primitive_atoms()
                    integration /= self.dynamic.structure.get_number_of_primitive_atoms()

            free_energy = thm.get_free_energy(temperature, frequency_range, power_spectrum_dos)
            entropy = thm.get_entropy(temperature, frequency_range, power_spectrum_dos)
            c_v = thm.get_cv(temperature, frequency_range, power_spectrum_dos)
            total_energy = thm.get_total_energy(temperature, frequency_range, power_spectrum_dos)

            power_spectrum_properties = [free_energy, entropy, c_v, total_energy, integration]
            print('\nThermal properties per unit cell ({0:.2f} K) [From DoS (resolution {1}) ]\n'
                  '-----------------------------------------------------------------'.format(temperature, resolution))
            print('                             Harmonic   Quasiparticle   Power spectrum\n')
            print('Free energy   (KJ/mol): {0:12.4f}  {5:12.4f}  {10:12.4f}\n'
                  'Entropy      (J/K/mol): {1:12.4f}  {6:12.4f}  {11:12.4f}\n'
                  'Cv           (J/K/mol): {2:12.4f}  {7:12.4f}  {12:12.4f}\n'
                  'Total energy  (KJ/mol): {3:12.4f}  {8:12.4f}  {13:12.4f}\n'
                  'Integration:            {4:12.4f}  {9:12.4f}  {14:12.4f}\n'.format(*(harmonic_properties +
                                                                                        renormalized_properties +
                                                                                        power_spectrum_properties)))
            if not self.parameters.silent:
                plt.plot(frequency_range, power_spectrum_dos, 'r-', label='Molecular dynamics')
                plt.axhline(y=0, color='k', ls='dashed')

        else:
            print('\nThermal properties per unit cell ({0:.2f} K) [From DoS (resolution {1})]\n'
                  '-----------------------------------------------------------------'.format(temperature, resolution))
            print('                            Harmonic    Quasiparticle\n')
            print('Free energy   (KJ/mol): {0:12.4f}  {5:12.4f}\n'
                  'Entropy      (J/K/mol): {1:12.4f}  {6:12.4f}\n'
                  'Cv           (J/K/mol): {2:12.4f}  {7:12.4f}\n'
                  'Total energy  (KJ/mol): {3:12.4f}  {8:12.4f}\n'
                  'Integration:            {4:12.4f}  {9:12.4f}\n'.format(
                *(harmonic_properties + renormalized_properties)))

        if not self.parameters.silent:
            self.plot_dos_phonopy(force_constants=self.get_renormalized_force_constants())

    def get_anisotropic_displacement_parameters(self, coordinate_type='uvrs', print_on_screen=True):

        elements = self.dynamic.structure.get_atomic_elements()

        atom_type = self.dynamic.structure.get_atom_type_index()
        atom_type_index_unique = np.unique(atom_type, return_index=True)[1]

        atom_equivalent = np.unique(atom_type, return_counts=True)[1]
        atomic_types_unique = [elements[i] for i in atom_type_index_unique]

        average_positions = self.dynamic.get_mean_displacement_matrix(use_average_positions=True)
        if print_on_screen:
            print('Anisotropic displacement parameters ({0}) [relative to average atomic positions]'.format(
                coordinate_type))
            print('          U11          U22          U33          U23          U13          U12')

        anisotropic_displacements = []
        for i, u_cart in enumerate(average_positions):

            cell = self.dynamic.structure.get_cell()
            cell_inv = np.linalg.inv(cell)  # Check this point
            n = np.array([[np.linalg.norm(cell_inv.T[0]), 0, 0],
                          [0, np.linalg.norm(cell_inv.T[1]), 0],
                          [0, 0, np.linalg.norm(cell_inv.T[2])]])

            u_crys = np.dot(np.dot(cell_inv.T, u_cart), cell_inv)
            u_uvrs = np.dot(np.dot(np.linalg.inv(n), u_crys), np.linalg.inv(n).T)

            u = {'cart': u_cart,
                 'crys': u_crys,
                 'uvrs': u_uvrs}

            if print_on_screen:
                for equivalent in range(atom_equivalent[i]):
                    print(
                        '{0:3} {1:12.8f} {2:12.8f} {3:12.8f} {4:12.8f} {5:12.8f} {6:12.8f}'.format(
                            atomic_types_unique[i],
                            u[coordinate_type][0, 0],
                            u[coordinate_type][1, 1],
                            u[coordinate_type][2, 2],
                            u[coordinate_type][1, 2],
                            u[coordinate_type][0, 2],
                            u[coordinate_type][0, 1]))

            anisotropic_displacements.append(u[coordinate_type])

        return anisotropic_displacements

    def get_average_atomic_positions(self, to_unit_cell=True):
        print('Average atomic positions')

        supercell = None
        if not to_unit_cell:
            supercell = self.dynamic.get_supercell_matrix()

        positions_average = self.dynamic.average_positions(to_unit_cell=to_unit_cell)
        elements = self.dynamic.structure.get_atomic_elements(supercell=supercell)

        for i, coordinate in enumerate(positions_average):
            print('{0:2} '.format(elements[i]) + '{0:15.8f} {1:15.8f} {2:15.8f}'.format(*coordinate.real))

    # QHA methods
    def set_qha_force_constants(self, fc_qha_file):
        self._force_constants_qha = pho_interface.get_force_constants_from_file(fc_qha_file,
                                                                                fc_supercell=self.dynamic.structure.get_supercell_phonon())

    def get_qha_shift(self, reduced_q_vector):
        if self._force_constants_qha is not None:
            import copy
            structure_qha = copy.copy(self.dynamic.structure)
            structure_qha.set_force_constants(self._force_constants_qha)
            qha_frequencies = pho_interface.obtain_eigenvectors_and_frequencies(structure_qha,
                                                                                reduced_q_vector,
                                                                                print_data=False)[1]
            return qha_frequencies - self.get_frequencies()
        else:
            return None


# Support functions
def _vector_in_list(vector_test_list, vector_full_list):
    for vector_test in vector_test_list:
        for i, vector_full in enumerate(vector_full_list):
            if (vector_full == vector_test).all():
                return i
    return 0


def replace_list(text_string):
    substitutions = {'GAMMA': r'$\Gamma$',
                     }

    for item in substitutions.items():
        text_string = text_string.replace(item[0], item[1])
    return text_string



###########################################################################################
# Script for evaluating configurations contained in an xyz file with a trained model
# Authors: Ilyes Batatia, Gregor Simm
# This program is distributed under the MIT License (see MIT.md)
###########################################################################################

import argparse
from typing import Dict

import ase.data
import ase.io
import numpy as np
import torch
from macer.externals.e3nn_legacy import o3

from macer.externals.mace_bundled import data
from macer.externals.mace_bundled.cli.convert_e3nn_cueq import run as run_e3nn_to_cueq
from macer.externals.mace_bundled.modules.utils import extract_invariant
from macer.externals.mace_bundled.tools import torch_geometric, torch_tools, utils


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument("--configs", help="path to XYZ configurations", required=True)
    parser.add_argument("--model", help="path to model", required=True)
    parser.add_argument("--output", help="output path", required=True)
    parser.add_argument(
        "--device",
        help="select device",
        type=str,
        choices=["cpu", "cuda"],
        default="cpu",
    )
    parser.add_argument(
        "--enable_cueq",
        help="enable cuequivariance acceleration",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--default_dtype",
        help="set default dtype",
        type=str,
        choices=["float32", "float64"],
        default="float64",
    )
    parser.add_argument("--batch_size", help="batch size", type=int, default=64)
    parser.add_argument(
        "--compute_stress",
        help="compute stress",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--compute_bec",
        help="compute BEC",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--return_contributions",
        help="model outputs energy contributions for each body order, only supported for MACE, not ScaleShiftMACE",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--return_descriptors",
        help="model outputs MACE descriptors",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--descriptor_num_layers",
        help="number of layers to take descriptors from",
        type=int,
        default=-1,
    )
    parser.add_argument(
        "--descriptor_aggregation_method",
        help="method for aggregating node features. None saves descriptors for each atom.",
        choices=["mean", "per_element_mean", None],
        default=None,
    )
    parser.add_argument(
        "--descriptor_invariants_only",
        help="save invariant (l=0) descriptors only",
        type=bool,
        default=True,
    )
    parser.add_argument(
        "--return_node_energies",
        help="model outputs MACE node energies",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--info_prefix",
        help="prefix for energy, forces and stress keys",
        type=str,
        default="MACE_",
    )
    parser.add_argument(
        "--head",
        help="Model head used for evaluation",
        type=str,
        required=False,
        default=None,
    )
    return parser.parse_args()


def get_model_output(
    model: torch.nn.Module,
    batch: Dict[str, torch.Tensor],
    compute_stress: bool,
    compute_bec: bool,
) -> Dict[str, torch.Tensor]:
    forward_args = {
        "compute_stress": compute_stress,
    }
    if compute_bec:
        # Only add `compute_bec` if it is requested
        # We check if the model is MACELES at the start of the run function
        forward_args["compute_bec"] = compute_bec
    return model(batch, **forward_args)


def main() -> None:
    args = parse_args()
    run(args)


def run(args: argparse.Namespace) -> None:
    torch_tools.set_default_dtype(args.default_dtype)
    device = torch_tools.init_device(args.device)

    # Load model
    model = torch.load(f=args.model, map_location=args.device)
    if model.__class__.__name__ != "MACELES" and args.compute_bec:
        raise ValueError("BEC can only be computed with MACELES model. ")
    if args.enable_cueq:
        print("Converting models to CuEq for acceleration")
        model = run_e3nn_to_cueq(model, device=device)
    model = model.to(
        args.device
    )  # shouldn't be necessary but seems to help with CUDA problems

    for param in model.parameters():
        param.requires_grad = False

    # Load data and prepare input
    atoms_list = ase.io.read(args.configs, index=":")
    if args.head is not None:
        for atoms in atoms_list:
            atoms.info["head"] = args.head
    configs = [data.config_from_atoms(atoms) for atoms in atoms_list]

    z_table = utils.AtomicNumberTable([int(z) for z in model.atomic_numbers])

    try:
        heads = model.heads
    except AttributeError:
        heads = None

    data_loader = torch_geometric.dataloader.DataLoader(
        dataset=[
            data.AtomicData.from_config(
                config, z_table=z_table, cutoff=float(model.r_max), heads=heads
            )
            for config in configs
        ],
        batch_size=args.batch_size,
        shuffle=False,
        drop_last=False,
    )

    # Collect data
    energies_list = []
    contributions_list = []
    descriptors_list = []
    node_energies_list = []
    stresses_list = []
    bec_list = []
    qs_list = []
    forces_collection = []

    for batch in data_loader:
        batch = batch.to(device)
        output = get_model_output(
            model, batch.to_dict(), args.compute_stress, args.compute_bec
        )
        energies_list.append(torch_tools.to_numpy(output["energy"]))
        if args.compute_stress:
            stresses_list.append(torch_tools.to_numpy(output["stress"]))
        if args.compute_bec:
            becs = np.split(
                torch_tools.to_numpy(output["BEC"]),
                indices_or_sections=batch.ptr[1:],
                axis=0,
            )
            bec_list.append(becs[:-1])  # drop last as its empty

            qs = np.split(
                torch_tools.to_numpy(output["latent_charges"]),
                indices_or_sections=batch.ptr[1:],
                axis=0,
            )
            qs_list.append(qs[:-1])  # drop last as its empty

        if args.return_contributions:
            contributions_list.append(torch_tools.to_numpy(output["contributions"]))

        if args.return_descriptors:
            num_layers = args.descriptor_num_layers
            if num_layers == -1:
                num_layers = int(model.num_interactions)
            irreps_out = o3.Irreps(str(model.products[0].linear.irreps_out))
            l_max = irreps_out.lmax
            num_invariant_features = irreps_out.dim // (l_max + 1) ** 2
            per_layer_features = [
                irreps_out.dim for _ in range(int(model.num_interactions))
            ]
            per_layer_features[-1] = (
                num_invariant_features  # Equivariant features not created for the last layer
            )

            descriptors = output["node_feats"]

            if args.descriptor_invariants_only:
                descriptors = extract_invariant(
                    descriptors,
                    num_layers=num_layers,
                    num_features=num_invariant_features,
                    l_max=l_max,
                )

            to_keep = np.sum(per_layer_features[:num_layers])
            descriptors = descriptors[:, :to_keep].detach().cpu().numpy()

            descriptors = np.split(
                descriptors,
                indices_or_sections=batch.ptr[1:],
                axis=0,
            )
            descriptors_list.extend(descriptors[:-1])  # drop last as its empty

        if args.return_node_energies:
            node_energies_list.append(
                np.split(
                    torch_tools.to_numpy(output["node_energy"]),
                    indices_or_sections=batch.ptr[1:],
                    axis=0,
                )[
                    :-1
                ]  # drop last as its empty
            )

        forces = np.split(
            torch_tools.to_numpy(output["forces"]),
            indices_or_sections=batch.ptr[1:],
            axis=0,
        )
        forces_collection.append(forces[:-1])  # drop last as its empty

    energies = np.concatenate(energies_list, axis=0)
    forces_list = [
        forces for forces_list in forces_collection for forces in forces_list
    ]
    assert len(atoms_list) == len(energies) == len(forces_list)
    if args.compute_stress:
        stresses = np.concatenate(stresses_list, axis=0)
        assert len(atoms_list) == stresses.shape[0]

    if args.compute_bec:
        bec_list = [becs for sublist in bec_list for becs in sublist]
        qs_list = [qs for sublist in qs_list for qs in sublist]

    if args.return_contributions:
        contributions = np.concatenate(contributions_list, axis=0)
        assert len(atoms_list) == contributions.shape[0]

    if args.return_descriptors:
        # no concatentation  - elements of descriptors_list have non-uniform shapes
        assert len(atoms_list) == len(descriptors_list)

    if args.return_node_energies:
        node_energies = np.concatenate(node_energies_list, axis=0)
        assert len(atoms_list) == node_energies.shape[0]

    # Store data in atoms objects
    for i, (atoms, energy, forces) in enumerate(zip(atoms_list, energies, forces_list)):
        atoms.calc = None  # crucial
        atoms.info[args.info_prefix + "energy"] = energy
        atoms.arrays[args.info_prefix + "forces"] = forces

        if args.compute_stress:
            atoms.info[args.info_prefix + "stress"] = stresses[i]

        if args.compute_bec:
            atoms.arrays[args.info_prefix + "BEC"] = bec_list[i].reshape(-1, 9)
            atoms.arrays[args.info_prefix + "latent_charges"] = qs_list[i]

        if args.return_contributions:
            atoms.info[args.info_prefix + "BO_contributions"] = contributions[i]

        if args.return_descriptors:
            descriptors = descriptors_list[i]
            if args.descriptor_aggregation_method:
                if args.descriptor_aggregation_method == "mean":
                    descriptors = np.mean(descriptors, axis=0)
                elif args.descriptor_aggregation_method == "per_element_mean":
                    descriptors = {
                        element: np.mean(
                            descriptors[atoms.symbols == element], axis=0
                        ).tolist()
                        for element in np.unique(atoms.symbols)
                    }
                atoms.info[args.info_prefix + "descriptors"] = descriptors
            else:  # args.descriptor_aggregation_method is None
                # Save descriptors for each atom (default behavior)
                atoms.arrays[args.info_prefix + "descriptors"] = np.array(descriptors)

        if args.return_node_energies:
            atoms.arrays[args.info_prefix + "node_energies"] = node_energies[i]

    # Write atoms to output path
    ase.io.write(args.output, images=atoms_list, format="extxyz")


if __name__ == "__main__":
    main()
